#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from rayvens.core import kamel_utils
from rayvens.core.mode import mode
from rayvens.core import catalog_utils


# Method to install kamel in a cluster.
# The cluster needs to be already started. An operator image, and a registry
# that Camel-K can use to publish newly created images to are needed.
def install(kamelImage,
            publishRegistry,
            mode=mode,
            localCluster=False,
            usingKind=False,
            insecureRegistry=False):
    # Enforce local cluster, for now.
    # TODO: make this work in an actual cluster.
    if not localCluster:
        raise RuntimeError('only local clusters are supported')

    # If running a local cluster (i.e. cluster running on local machine)
    # then only the kind cluster is supported.
    # TODO: enable this for other cluster types.
    if localCluster and not usingKind:
        raise RuntimeError('locally, only kind cluster is supported')

    command = ["install"]

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    # Add kamel operator image.
    command.append("--operator-image")
    command.append(kamelImage)

    # Add registry that kamel can use for image publication.
    command.append("--registry")

    # Kind cluster starts with a registry.
    if usingKind:
        command.append(publishRegistry)

    # Local registry used to publish newly constructed images is insecure.
    # TODO: support secure registries.
    if (insecureRegistry):
        command.append("--registry-insecure")

    # Force installation.
    command.append("--force")

    return kamel_utils.invoke_kamel_command(command, mode, "camel-k-operator")


# Invoke kamel uninstall.
def uninstall(install_invocation):
    command = ["uninstall"]

    # Namespace
    command.append("-n")
    command.append(install_invocation.mode.namespace)

    return kamel_utils.invoke_kamel_command(command, install_invocation.mode,
                                            "camel-k-operator")


# Kamel run invocation.
def run(integration_content,
        mode,
        integration_name,
        envVars=[],
        integration_type=None):
    command = ["run"]

    # Append ProcessPath.java file.
    if catalog_utils.integration_requires_path_processor(integration_content):
        process_file = os.path.join(os.path.dirname(__file__),
                                    'ProcessPath.java')
        command.append(process_file)

    # Append ProcessFile.java file.
    if catalog_utils.integration_requires_file_processor(integration_content):
        process_file = os.path.join(os.path.dirname(__file__),
                                    'ProcessFile.java')
        command.append(process_file)

    # Use appropriate Queue type(s).
    if mode.transport == 'http' and integration_type == 'source':
        if catalog_utils.integration_requires_file_queue(integration_content):
            queue = os.path.join(os.path.dirname(__file__), 'FileQueue.java')
            command.append(queue)

        if catalog_utils.integration_requires_file_queue_json(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'FileQueueJson.java')
            command.append(queue)

            command.append("-d")
            command.append("mvn:com.googlecode.json-simple:json-simple:1.1.1")

        if catalog_utils.integration_requires_file_watch_queue(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'FileWatchQueue.java')
            command.append(queue)

            command.append("-d")
            command.append("mvn:com.googlecode.json-simple:json-simple:1.1.1")

        if catalog_utils.integration_requires_meta_event_queue(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'MetaEventQueue.java')
            command.append(queue)

        if catalog_utils.integration_requires_queue(integration_content):
            queue = os.path.join(os.path.dirname(__file__), 'Queue.java')
            command.append(queue)

    # Integration name.
    command.append("--name")
    command.append(integration_name)

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    # Include health check:
    command.append("-d")
    command.append("camel:camel-quarkus-microprofile-health")

    for envVar in envVars:
        if envVar not in os.environ:
            raise RuntimeError("Variable %s not set in current environment" %
                               envVar)
        command.append("--env")
        command.append("%s=%s" % (envVar, os.getenv(envVar)))

    return kamel_utils.invoke_kamel_command(
        command,
        mode,
        integration_name,
        integration_content=integration_content)


# Kamel local run invocation.
def local_run(integration_content,
              mode,
              integration_name,
              envVars=[],
              port=None,
              integration_type=None):
    command = ["local", "run"]

    # Append ProcessPath.java file.
    if catalog_utils.integration_requires_path_processor(integration_content):
        process_file = os.path.join(os.path.dirname(__file__),
                                    'ProcessPath.java')
        command.append(process_file)

    # Append ProcessFile.java file.
    if catalog_utils.integration_requires_file_processor(integration_content):
        process_file = os.path.join(os.path.dirname(__file__),
                                    'ProcessFile.java')
        command.append(process_file)

    # Use appropriate Queue type(s).
    if mode.transport == 'http' and integration_type == 'source':
        if catalog_utils.integration_requires_file_queue(integration_content):
            queue = os.path.join(os.path.dirname(__file__), 'FileQueue.java')
            command.append(queue)

        if catalog_utils.integration_requires_file_queue_json(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'FileQueueJson.java')
            command.append(queue)

            command.append("-d")
            command.append("mvn:com.googlecode.json-simple:json-simple:1.1.1")

        if catalog_utils.integration_requires_file_watch_queue(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'FileWatchQueue.java')
            command.append(queue)

            command.append("-d")
            command.append("mvn:com.googlecode.json-simple:json-simple:1.1.1")

        if catalog_utils.integration_requires_meta_event_queue(
                integration_content):
            queue = os.path.join(os.path.dirname(__file__),
                                 'MetaEventQueue.java')
            command.append(queue)

            command.append("-d")
            command.append("mvn:com.googlecode.json-simple:json-simple:1.1.1")

        if catalog_utils.integration_requires_queue(integration_content):
            queue = os.path.join(os.path.dirname(__file__), 'Queue.java')
            command.append(queue)

    # Port is mandatory.
    if port is None:
        return RuntimeError('port is missing')

    # In the case of HTTP, add the port:
    command.append('--property')
    command.append(f'quarkus.http.port={port}')

    # Include health check:
    command.append("-d")
    command.append("camel:camel-quarkus-microprofile-health")

    return kamel_utils.invoke_kamel_command(
        command,
        mode,
        integration_name,
        integration_content=integration_content)


# Kamel log command (needs operator).
def log(mode, integration_name, custom_message):
    command = ["log"]

    # Add integration name.
    command.append(integration_name)

    # Namespace
    command.append("-n")
    command.append(mode.namespace)

    return kamel_utils.invoke_kamel_command(command,
                                            mode,
                                            integration_name,
                                            message=custom_message)


# Kamel delete invocation (needs operator).
def delete(integration_invocation, integration_name):
    # Compose command with integration name.
    command = ["delete"]

    # Namespace
    command.append("-n")
    command.append(integration_invocation.mode.namespace)

    command.append(integration_name)

    return kamel_utils.invoke_kamel_command(command,
                                            integration_invocation.mode,
                                            integration_name)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import subprocess
import os
import signal
import yaml
import sys
import time
from rayvens.core import utils
from rayvens.core import kamel_utils
from rayvens.core import kubernetes_utils
from rayvens.cli.utils import PRINT, VERBOSE_MODE

#
# Kamel invocation.
#


class KamelInvocation:
    subprocess_name = "Kamel"

    def __init__(self,
                 command_options,
                 mode,
                 integration_name="",
                 integration_content=[]):
        self.mode = mode
        self.integration_name = integration_name

        # Get subcommand type.
        self.subcommand_type = kamel_utils.kamel_command_type(command_options)

        # Use an external harness process to handle graceful termination of
        # integrations.
        harness = os.path.join(os.path.dirname(__file__), 'harness.py')
        final_command = [sys.executable, harness]
        if self.subcommand_type == kamel_utils.KamelCommand.LOCAL_RUN:
            final_command.append('kamel')
        elif self.subcommand_type == kamel_utils.KamelCommand.RUN:
            final_command.extend(
                [self.integration_name, self.mode.namespace, 'kamel'])
        else:
            final_command = ['kamel']
        final_command.extend(command_options)

        # If integration content is not null then we have files to create and
        # write to.
        # TODO: support multiple contents.
        self.filename = None
        for file_content in integration_content:
            self.filename = f'{self.integration_name}.yaml'
            with open(self.filename, 'w') as f:
                yaml.dump(file_content, f)
            self.filename = os.path.abspath(self.filename)
            final_command.append(self.filename)

        # Log kamel command.
        print("Exec command => ", " ".join(final_command))

        # Add to PATH for case when this command is invoked in a cluster.
        if mode.is_cluster():
            os.environ['PATH'] = ":".join(
                ["/home/ray/rayvens/rayvens/linux-x86_64",
                 os.getenv('PATH')])

        # Fail early before command is invoked if kamel is not found.
        if self.uses_operator() and not utils.executable_is_available("kamel"):
            raise RuntimeError('kamel executable not found in PATH')

        # Get end condition or fail if command type is not supported.
        self.end_condition = kamel_utils.kamel_command_end_condition(
            self.subcommand_type, self.integration_name)

        # Launch kamel command in a new process.
        self.process = subprocess.Popen(final_command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        start_new_session=True)

    def invoke(self, message):
        if message is None:
            return self._check_command_outcome()

        return self._check_logs(message, with_timeout=True)

    def runs_integration(self):
        return self.subcommand_type == kamel_utils.KamelCommand.RUN or \
            self.subcommand_type == kamel_utils.KamelCommand.LOCAL_RUN

    def uses_operator(self):
        return self.subcommand_type == kamel_utils.KamelCommand.RUN or \
            self.subcommand_type == kamel_utils.KamelCommand.LOG or \
            self.subcommand_type == kamel_utils.KamelCommand.INSTALL or \
            self.subcommand_type == kamel_utils.KamelCommand.DELETE

    def kill(self):
        # Magic formula for terminating all processes in the group including
        # any subprocesses that the kamel command might have created.
        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

    def cancel(self):
        try:
            os.kill(self.process.pid, signal.SIGTERM)
        except ProcessLookupError:
            pass

    def _check_command_outcome(self):
        # Check if kamel instance launched correctly.
        success = self._check_kamel_output(self.end_condition,
                                           with_output=True)

        # Emit success/fail message.
        subcommand = kamel_utils.kamel_command_str(self.subcommand_type)
        log = "Kamel `%s` command finished successfully." % subcommand
        if not success:
            log = "Kamel `%s` command failed." % subcommand
        utils.print_log(self.subprocess_name, log)

        # Delete intermediate file.
        if self.filename is not None:
            os.remove(self.filename)

        return success

    def _check_logs(self, message, with_timeout=False):
        # Check logs.
        success = self._check_kamel_output(message,
                                           with_output=True,
                                           with_timeout=with_timeout)

        # Emit success/fail message.
        log = "Logs checked successfully."
        if not success:
            log = "Log check failed or timed out."
        utils.print_log(self.subprocess_name, log)
        return success

    def _check_kamel_output(self,
                            end_condition,
                            with_output=False,
                            with_timeout=False):
        # Decide whether to print logs or not. If this is release mode
        # then no output will be printed:
        with_output_final = with_output
        if self.mode.release:
            with_output_final = False

        # Implicit 5 minute timout in hundredths of a second:
        timout_duration = 5 * 60 * 100
        if with_timeout:
            countdown = timout_duration
            reading_thread = utils.LogThread(self.process.stdout)

            # Kill thread when program ends in case it does not end before
            # that.
            reading_thread.daemon = True

            # Start thread analyzing logs:
            reading_thread.start()

            # Initiate the first read:
            reading_thread.read_flag.set()

        success = False
        while True:
            # Log progress of kamel subprocess:
            if with_timeout:
                output = utils.print_log_from_queue(self.subprocess_name,
                                                    reading_thread.queue,
                                                    with_output_final)
            else:
                output = utils.print_log_from_subprocess(
                    self.subprocess_name, self.process.stdout,
                    with_output_final)

            # Use the Kamel output to decide when Kamel instance is
            # ready to receive requests.
            if output is not None:
                if with_timeout:
                    countdown = timout_duration
                if end_condition in output:
                    success = True
                    break

            # Check process has not exited prematurely.
            if self.process.poll() is not None:
                break

            # If timeout is enabled we decrement countdown.
            if with_timeout:
                countdown -= 1
                if countdown == 0:
                    break
                if output is not None:
                    reading_thread.read_flag.set()
                time.sleep(0.01)

        # Terminate log thread:
        if with_timeout:
            reading_thread.stop_flag.set()
            if success:
                reading_thread.join()

        return success


#
# Handle calls to Kubernetes kubectl.
#


class KubectlInvocation:
    subprocess_name = "Kubectl"

    def __init__(self, command_options, service_name=""):
        # If list is porvided, join it.
        if isinstance(command_options, list):
            command_options = " ".join(command_options)

        # Initialize state.
        self.subcommand_type = kubernetes_utils.kubectl_command_type(
            command_options)
        self.pod_name = ""

        # Get end condition or fail if command type is not supported.
        self.end_condition = kubernetes_utils.kamel_command_end_condition(
            self.subcommand_type, service_name)

        # Create the kubectl command.
        command = " ".join(["exec", "kubectl", command_options])

        if not utils.executable_is_available("kubectl"):
            raise RuntimeError('kubectl executable not found in PATH')

        # Log kamel command.
        print("Exec command => ", command)

        # Launch kamel command in a new process.
        self.process = subprocess.Popen(command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        shell=True,
                                        preexec_fn=os.setsid)

    def invoke(self, message, service_name, with_output=False):
        # Set end condition for custom message if one is provided.
        end_condition = self.end_condition
        if message is not None:
            end_condition = message

        # Check command output.
        success = self._check_kubectl_output(end_condition,
                                             service_name=service_name,
                                             with_output=with_output)

        subcommand = kubernetes_utils.kubectl_command_str(self.subcommand_type)
        log = "Kubectl %s command finished successfully." % subcommand
        if not success:
            log = "Kubectl %s command failed." % subcommand
        utils.print_log(self.subprocess_name, log)

        return success

    def pod_is_running(self, integration_name):
        running = False
        while True:
            # Process output line by line until we find the pod we are looking
            # for.
            # There should only be one new pod.
            output = utils.print_log_from_subprocess(self.subprocess_name,
                                                     self.process.stdout, True)
            if self.pod_name == "":
                self.pod_name = kubernetes_utils.extract_pod_name(
                    output, integration_name)

            if kubernetes_utils.is_pod_state_running(output, self.pod_name):
                running = True
                break

            if kubernetes_utils.is_pod_state_error(output, self.pod_name):
                break

            # Return if command has exited.
            if self.process.poll() is not None:
                break

        log = "Pod with name `%s` is now Running." % self.pod_name
        if not running:
            log = "Pod with name `%s` failed to start." % self.pod_name
        utils.print_log(self.subprocess_name, log)

        return running

    def kill(self):
        # Terminating all processes in the group including any subprocesses
        # that the kubectl command might have created.
        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

    def _check_kubectl_output(self,
                              end_condition,
                              service_name=None,
                              with_output=False):
        success = False
        while True:
            output = utils.print_log_from_subprocess(self.subprocess_name,
                                                     self.process.stdout,
                                                     with_output)
            if self.subcommand_type == \
               kubernetes_utils.KubectlCommand.GET_SERVICES:
                if kubernetes_utils.service_name_matches(output, service_name):
                    success = True
                    break
            elif end_condition in output:
                success = True
                break

            if self.process.poll() is not None:
                break
        return success


#
# Kafka topic creator invocation.
#


class KafkaInvocation:
    subprocess_name = "Kafka"

    def __init__(self, command_options):
        final_command = ['kafka-topics']
        final_command.extend(command_options)

        # Log kamel command.
        print("Exec command => ", " ".join(final_command))

        # Launch kafka command in a new process.
        self.process = subprocess.Popen(final_command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        start_new_session=True)

    def invoke(self, checked_topic):
        return self._check_kafka_output(checked_topic)

    def _check_kafka_output(self, checked_topic):
        while True:
            # Log progress of kafka topic creation subprocess:
            output = utils.print_log_from_subprocess(self.subprocess_name,
                                                     self.process.stdout, True)
            if checked_topic is not None and checked_topic in output:
                return True

            if checked_topic is None and "Created topic" in output:
                return True

            if checked_topic is None and "already exists" in output:
                return True

            # Check process has not exited prematurely.
            if self.process.poll() is not None:
                break

        return False


#
# Docker run invocation.
#


class DockerIntegrationInvocation:
    subprocess_name = "Docker"

    def __init__(self, command_options):
        final_command = ['docker']
        final_command.extend(command_options)

        # Log kamel command.
        command_to_print = " ".join(final_command)
        PRINT(f"Exec command => {command_to_print}", tag=self.subprocess_name)

        # Launch kafka command in a new process.
        self.process = subprocess.Popen(final_command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        start_new_session=True)

    def invoke(self, stop_check=None):
        return self._check_kafka_output(stop_check)

    def _check_kafka_output(self, stop_check):
        while True:
            # Log progress of kafka topic creation subprocess:
            output = utils.print_log_from_subprocess(self.subprocess_name,
                                                     self.process.stdout,
                                                     VERBOSE_MODE())
            if stop_check is not None and stop_check in output:
                return True

            # Check process has not exited prematurely.
            if self.process.poll() is not None:
                break

        return False



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ray
import requests
from ray import serve


class SinkEvent:
    def __init__(self, route, integration_name):
        self.route = route
        self.integration_name = integration_name
        self.data = None

    def get_data(self):
        if self.data is None:
            raise RuntimeError(
                "Attempting to send event with None data field.")
        return self.data


class KamelEventHandler:
    def __init__(self, mode, topic):
        self.mode = mode
        self.topic = topic

    async def __call__(self, request):
        body = await request.body()
        if isinstance(body, SinkEvent):
            endpoint = self.mode._get_server_address(
                body.integration_name) + body.route
            requests.post(endpoint, data=body.get_data())
            return {"message": "Success"}

        if self.topic is None:
            return {"message": "Failure"}
        print("Body:", body)
        self.topic.append.remote(body)
        return {"message": "Success"}


class KamelBackend:
    backendName = "kamel_backend"

    def __init__(self, mode, topic=None):
        # When the backend runs on a local machine we must allocate its
        # actor at least 1 CPU.
        # actor_options = {'num_cpus': 0}
        # if mode.is_local() or mode.is_mixed():
        #     actor_options = {'num_cpus': 1}
        serve.create_backend(self.backendName, KamelEventHandler, mode, topic)
        self.endpoint_to_event = {}

    def createProxyEndpoint(self, endpoint_name, route, integration_name):
        self.endpoint_to_event[endpoint_name] = SinkEvent(
            route, integration_name)

        # Create endpoint with method as POST.
        serve.create_endpoint(endpoint_name,
                              backend=self.backendName,
                              route=route,
                              methods=["POST"])

    def _post_event(self, endpointHandle, endpoint_name, data):
        # Get partial event.s
        event = self.endpoint_to_event[endpoint_name]

        # Populate data field.
        event.data = data

        # Send request to backend.
        return ray.get(endpointHandle.remote(event))

    def postToProxyEndpoint(self, endpoint_name, data):
        return self._post_event(serve.get_handle(endpoint_name), endpoint_name,
                                data)

    def postToProxyEndpointHandle(self, endpointHandle, endpoint_name, data):
        return self._post_event(endpointHandle, endpoint_name, data)

    def removeProxyEndpoint(self, endpoint_name):
        serve.delete_endpoint(endpoint_name)
        self.endpoint_to_event.pop(endpoint_name)


# Method which send post request to external Camel-K sink.


@ray.remote
class SinkSubscriber(object):
    def __init__(self, route):
        self.route = route

    def sendToSink(self, data):
        requests.post("http://0.0.0.0:8080" + self.route, data=data)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import yaml
import rayvens.cli.utils as utils
import rayvens.cli.file as file
import rayvens.cli.docker as docker
import rayvens.cli.kubernetes as kube
import rayvens.core.utils as rayvens_utils
from rayvens.core.catalog import sources, sinks
from rayvens.core.catalog import construct_source, construct_sink
from rayvens.cli.docker import docker_run_integration
run_tag = "run"


def run_integration(args):
    # Form full image name:
    if args.image is None:
        raise RuntimeError("Missing image name")
    image = utils.get_integration_image(args)

    # Set verbosity:
    utils.verbose = args.verbose

    # Get a free port:
    docker.free_port = rayvens_utils.random_port()

    # Create a work directory in the current directory:
    workspace_directory = file.Directory("workspace")

    # Fetch summary file from integration image.
    docker.add_summary_from_image(image, workspace_directory)

    # Retrieve a reference to the summary file.
    summary_file = workspace_directory.get_file(file.summary_file_name)

    # Get integration kind from summary file:
    kind = summary_file.kind

    # Check if a valid launch image name has been passed:
    with_job_launcher = summary_file.launch_image != "None"

    # The name of the integration:
    name = kind
    if args.name is not None:
        name = args.name

    # Check if the source/sink is predefined.
    predefined_integration = kind in sources or kind in sinks

    # By default the HTTP transport is used. This is the only supported
    # transport for now.
    inverted_transport = True

    is_sink = False
    if predefined_integration:
        # Extract predefined integration kind:
        full_config = utils.get_full_config(summary_file, args)

        # Create the integration yaml specification.
        route = "/" + name + "-route"
        if kind in sources:
            spec = construct_source(full_config,
                                    f'platform-http:{route}',
                                    inverted=inverted_transport)
        else:
            spec = construct_sink(full_config, f'platform-http:{route}')
            is_sink = True

        # Write the specification to the file.
        modeline_options = utils.get_modeline_config(args, summary_file)
        integration_source_file = modeline_options + "\n\n" + yaml.dump(spec)
        integration_file = file.File(
            utils.get_kubernetes_integration_file_name(name),
            contents=integration_source_file)
    else:
        raise RuntimeError("Not implemented yet")

    # Fetch the variables specified as environment variables.
    envvars = utils.get_modeline_envvars(summary_file, args)

    server_address = None
    integration_file.emit()
    if args.deploy is not None and args.deploy:
        # Set the namespace:
        namespace = "default"
        if args.namespace is not None:
            namespace = args.namespace

        # Update integration file on image via configMap:
        integration_config_map = kube.ConfigMap(integration_file,
                                                namespace=namespace)

        # Prepare Kubernetes API:
        from kubernetes import client, config
        config.load_kube_config()

        # Kubernetes client:
        k8s_client = client.ApiClient()

        # Create configMap that updates the integration file.
        integration_config_map.create(k8s_client)

        # Deploy integration in Kubernetes:
        deployment = kube.get_deployment(name, namespace,
                                         utils.get_registry(args), args,
                                         with_job_launcher,
                                         integration_config_map, k8s_client)

        deployment.create(k8s_client)
    else:
        # Output endpoint for sink:
        server_address = f"http://localhost:{docker.free_port}"

        # Run final integration image:
        #   docker run \
        #      -v integration_file_path:/workspace/<integration_file_name> \
        #      --env ENV_VAR=$ENV_VAR -p <free_host_port>:8080 \
        #      <image>
        docker_run_integration(image,
                               integration_file.full_path,
                               integration_file.name,
                               name,
                               envvars=envvars,
                               is_sink=is_sink,
                               server_address=server_address)

    integration_file.delete()

    if is_sink and server_address is not None:
        print(f"{server_address}{route}")



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import yaml


def http_source():
    return dict(required=['url'], optional=['period'])


def kafka_source():
    return dict(required=['topic', 'brokers'],
                optional=['SASL_password', 'JAAS_config'])


def telegram_source():
    return dict(required=['authorization_token'], optional=[])


def binance_source():
    return dict(required=['coin'], optional=['period'])


def cos_source():
    return dict(required=[
        'bucket_name', 'access_key_id', 'secret_access_key', 'endpoint'
    ],
                optional=['region', 'move_after_read', 'meta_event_only'])


def file_source():
    return dict(required=['path'], optional=['keep_files', 'recursive'])


def file_watch_source():
    return dict(required=['path'], optional=['events', 'recursive'])


def generic_source():
    return dict(required=[], optional=[], pair=[('uri', 'spec')])


def generic_periodic_source():
    return dict(required=[], optional=['period'], pair=[('uri', 'spec')])


source_requirements = {
    'http-source': http_source,
    'kafka-source': kafka_source,
    'telegram-source': telegram_source,
    'binance-source': binance_source,
    'cloud-object-storage-source': cos_source,
    'file-source': file_source,
    'file-watch-source': file_watch_source,
    'generic-source': generic_source,
    'generic-periodic-source': generic_periodic_source
}


def slack_sink():
    return dict(required=['channel', 'webhook_url'], optional=[])


def kafka_sink():
    return dict(required=['topic', 'brokers'], optional=['SASL_password'])


def telegram_sink():
    return dict(required=['authorization_token'], optional=['chat_id'])


def cos_sink():
    return dict(required=[
        'bucket_name', 'access_key_id', 'secret_access_key', 'endpoint'
    ],
                optional=[
                    'region', 'file_name', 'from_file', 'from_directory',
                    'upload_type', 'batch_size', 'messages_per_batch',
                    'part_size', 'keep_file'
                ])


def generic_sink():
    return dict(required=[], optional=[], pair=[('uri', 'spec')])


sink_requirements = {
    'slack-sink': slack_sink,
    'kafka-sink': kafka_sink,
    'telegram-sink': telegram_sink,
    'cloud-object-storage-sink': cos_sink,
    'generic-sink': generic_sink
}


def integration_requirements(kind):
    handler = source_requirements.get(kind)
    if handler is None:
        handler = sink_requirements.get(kind)
        if handler is None:
            raise TypeError(f'Unsupported Camel source or sink: {kind}.')
    return handler()


def print_predefined_integrations():
    print(" Predefined sources:")
    for integration_type in source_requirements:
        print(f"     {integration_type}")
    print(" Predefined sinks:")
    for integration_type in sink_requirements:
        print(f"     {integration_type}")


def print_requirements_summary(kind):
    requirements = integration_requirements(kind)

    print(f" Integration: {kind}")
    if len(requirements['required']) > 0:
        print("    Required properties:")
        for req_property in requirements['required']:
            print(f"        {req_property}")
    if len(requirements['optional']) > 0:
        print("    Optional properties:")
        for opt_property in requirements['optional']:
            print(f"        {opt_property}")
    if 'pair' in requirements and len(requirements['pair']) > 0:
        print("    At least one of the properties is required:")
        for pair_property in requirements['pair']:
            print(f"        {pair_property}")


def get_all_properties(kind):
    all_properties = []
    requirements = integration_requirements(kind)
    all_properties.extend(requirements['required'])
    all_properties.extend(requirements['optional'])
    if 'pair' in requirements:
        for pair in requirements['pair']:
            all_properties.append(pair.first)
            all_properties.append(pair.second)
    return all_properties


source_modeline_name = {
    'http-source': "http",
    'kafka-source': "kafka",
    'telegram-source': "telegram",
    'binance-source': "xchange.binance",
    'cloud-object-storage-source': "aws2-s3",
    'file-source': "file",
    'file-watch-source': "file-watch"
}

sink_modeline_name = {
    'slack-sink': "slack",
    'kafka-sink': "kafka",
    'telegram-sink': "telegram",
    'cloud-object-storage-sink': "aws2-s3"
}


def get_modeline_component_name(kind):
    component_name = None
    if kind in source_modeline_name:
        component_name = source_modeline_name.get(kind)
    if kind in sink_modeline_name:
        component_name = sink_modeline_name.get(kind)
    if component_name is None:
        raise TypeError(f'Modeline unsupported for source or sink: {kind}.')
    return "camel.component." + component_name


modeline_property_name_exceptions = {
    'SASL_password': 'sslKeyPassword',
    'JAAS_config': 'saslJaasConfig',
    'authorization_token': 'authorizationToken',
    'access_key_id': 'accessKey',
    'secret_access_key': 'secretKey',
    'endpoint': 'uriEndpointOverride',
    'move_after_read': 'destinationBucket',
    'webhook_url': 'webhookUrl',
    'batch_size': 'batchSize',
    'messages_per_batch': 'batchMessageNumber',
    'part_size': 'partSize',
}


def get_modeline_property_name(property_name):
    if property_name in modeline_property_name_exceptions:
        return modeline_property_name_exceptions.get(property_name)

    if property_name in cannot_be_modeline_property:
        raise TypeError(
            f'Cannot pass property {property_name} via environment variable.')
    return property_name


cannot_be_modeline_property = [
    'url', 'topic', 'channel', 'coin', 'period', 'meta_event_only', 'path',
    'keep_files', 'events', 'chat_id', 'bucket_name', 'file_name', 'from_file',
    'from_directory', 'upload_type'
]


def can_be_modeline_property(property_name):
    return property_name not in cannot_be_modeline_property


def get_modeline_properties(kind, envvars):
    # We assume at this point that:
    # 1. All property env var pairs are syntactically correct
    # 2. The property env var pairs do not contain properties that
    #    cannot be modelined.
    component_name = get_modeline_component_name(kind)

    modeline_properties = {}
    for property_env_pair in envvars:
        components = property_env_pair.split("=")
        property_name = components[0]
        env_var_name = "{{env:" + components[1] + "}}"

        # Modeline property format:
        # # camel-k: property=<property_name>={{env:<env_var>}}
        modeline_property = get_modeline_property_name(property_name)
        modeline_property = ".".join([component_name, modeline_property])
        modeline_property = "=".join(
            ["property", modeline_property, env_var_name])
        modeline_properties[property_name] = "# camel-k: " + modeline_property
    return modeline_properties


def fill_config(kind, property_value_list, show_missing=True):
    requirements = integration_requirements(kind)

    config = dict(kind=kind)

    # Assume list is of the form `property=value`
    for property_value in property_value_list:
        if "=" not in property_value:
            raise RuntimeError(
                f"Invalid property-value pair: {property_value}")

        components = property_value.split("=")
        config[components[0]] = "=".join(components[1:])

    # Check all requirements have been fulfilled.
    missing_requirements = []
    if len(requirements['required']) > 0:
        for req_property in requirements['required']:
            if req_property not in config:
                missing_requirements.append(req_property)

    if show_missing and len(missing_requirements) > 0:
        print(f" Missing required properties for integration type {kind}:")
        for missing_req in missing_requirements:
            print(f"    {missing_req}")

    return config, missing_requirements


def integration_requires_path_processor(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: processPath' in configuration:
        return True
    return False


def integration_requires_file_processor(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: processFile' in configuration:
        return True
    return False


def integration_requires_file_queue(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: addToFileQueue' in configuration:
        return True
    return False


def integration_requires_file_queue_json(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: addToFileJsonQueue' in configuration:
        return True
    return False


def integration_requires_file_watch_queue(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: addToFileWatchQueue' in configuration:
        return True
    return False


def integration_requires_meta_event_queue(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: addToMetaEventQueue' in configuration:
        return True
    return False


def integration_requires_queue(integration_content):
    configuration = yaml.dump(integration_content)
    if 'bean: addToQueue' in configuration:
        return True
    return False



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import subprocess
import rayvens.cli.version as v
import rayvens.cli.file as file
from rayvens.core.invocation import DockerIntegrationInvocation
from rayvens.cli.utils import PRINT, VERBOSE_MODE
from rayvens.core.common import wait_for_ready_integration

image_workspace_name = "workspace"
local_workspace_name = "docker_workspace"
built_integration_directory = "my-integration"
input_port_var_name = "INPUT_PORT"
docker_tag = "Docker"
free_port = None


def docker_run_integration(image_name,
                           local_integration_path,
                           integration_file_name,
                           name,
                           envvars=[],
                           is_sink=False,
                           server_address=None):
    # Build command:
    command = ["run"]

    # Mount integration file:
    # -v local_integration_path:inside_image_integration_path
    inside_image_integration_path = get_file_on_image(integration_file_name)

    # Mount argument:
    local_to_image_integration_mount = ":".join(
        [str(local_integration_path), inside_image_integration_path])

    if local_to_image_integration_mount is not None:
        command.append("-v")
        command.append(local_to_image_integration_mount)

    # Add env vars:
    for envvar in envvars:
        command.append("--env")
        command.append(f"{envvar}=${envvar}")

    # Add communication port:
    if is_sink:
        command.append("-p")
        command.append(f"{free_port}:8080")

    # Add container name:
    command.append("--name")
    command.append(get_unique_container_name(name))

    # Add image:
    command.append(image_name)

    # Invoke integration:
    docker_run_invocation = DockerIntegrationInvocation(command)
    success = docker_run_invocation.invoke(stop_check="Installed features")
    if success:
        PRINT(f"Image {image_name} has been run successfully.", tag=docker_tag)
    else:
        PRINT(f"Image {image_name} run failed.", tag=docker_tag)

    # Check integration is ready to receive events:
    if server_address is not None:
        integration_ready = wait_for_ready_integration(server_address)
        if integration_ready:
            PRINT(f"Integration inside {image_name} is ready.", tag=docker_tag)
        else:
            PRINT(f"Integration inside {image_name} failed.", tag=docker_tag)


def docker_build(working_directory, image_name):
    command = ["docker"]

    # Build command:
    command.append("build")

    # Path to working directory:
    command.append(working_directory)

    # Image name:
    command.append("-t")
    command.append(image_name)

    # Wait for docker command to finish before returning:
    outcome = subprocess.run(command)

    if outcome.returncode == 0:
        PRINT(f"Base image {image_name} built successfully.", tag=docker_tag)
    else:
        PRINT(f"Base image {image_name} building failed.", tag=docker_tag)


def docker_push(image_name):
    command = ["docker"]

    # Build command:
    command.append("push")

    # Image name:
    command.append(image_name)

    # Wait for docker command to finish before returning:
    outcome = subprocess.run(command)

    if outcome.returncode == 0:
        PRINT(f"Base image {image_name} pushed successfully.", tag=docker_tag)
    else:
        PRINT(f"Base image {image_name} push failed.", tag=docker_tag)


def docker_cp_to_host(container_id, from_file, to_file):
    command = ["docker"]

    # Copy command:
    command.append("cp")

    # Source:
    command.append(f"{container_id}:{from_file}")

    # Destination:
    command.append(f"{to_file}")

    # Wait for docker command to finish before returning:
    if VERBOSE_MODE():
        outcome = subprocess.run(command)
    else:
        outcome = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

    if outcome.returncode == 0:
        PRINT(f"Container file {from_file} successfully copied to host.",
              tag=docker_tag)
    else:
        PRINT(f"Container file {from_file} copy to host failed.",
              tag=docker_tag)


def docker_create(image):
    command = ["docker"]

    # Copy command:
    command.append("create")

    # Source:
    command.append(image)

    # Wait for docker command to finish before returning:
    if VERBOSE_MODE():
        outcome = subprocess.run(command, stdout=subprocess.PIPE)
    else:
        outcome = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

    if outcome.returncode == 0:
        PRINT(f"Container for {image} created successfully.", tag=docker_tag)
    else:
        PRINT(f"Container creation for {image} failed.", tag=docker_tag)

    return outcome.stdout.decode('utf8').strip()


def docker_kill(container_id):
    command = ["docker", "kill", container_id]
    if VERBOSE_MODE():
        outcome = subprocess.run(command)
    else:
        outcome = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

    if outcome.returncode == 0:
        PRINT(f"Container {container_id} killed successfully.", tag=docker_tag)
    else:
        PRINT(f"Container {container_id} kill failed.", tag=docker_tag)


def docker_rm(container_id):
    command = ["docker", "rm", container_id]
    if VERBOSE_MODE():
        outcome = subprocess.run(command)
    else:
        outcome = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

    if outcome.returncode == 0:
        PRINT(f"Container {container_id} removed successfully.",
              tag=docker_tag)
    else:
        PRINT(f"Container {container_id} removal failed.", tag=docker_tag)


def docker_container_ls():
    command = ["docker", "container", "ls", "--all"]

    container_fields = []
    container_fields.append("{{.Names}}")
    container_fields.append("{{.ID}}")

    if len(container_fields) > 0:
        command.append("--format")
        command.append("=".join(container_fields))

    # List all local containers:
    outcome = subprocess.run(command,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    if outcome is None:
        return {}

    container_pairs = outcome.stdout.decode('utf8').strip().split()

    containers = {}
    for container_pair in container_pairs:
        components = container_pair.split("=")
        containers[components[0]] = components[1]
    return containers


def get_unique_container_name(name):
    containers = docker_container_ls()
    new_name = name
    count = 0
    while True:
        found = False
        for container_name in containers:
            if new_name == container_name:
                new_name = name + "-" + str(count)
                count += 1
                found = True
                break
        if not found:
            return new_name


class DockerCopyStep:
    def __init__(self, source, destination, from_stage, from_image):
        self.kind = "COPY"
        self.source = source
        self.destination = destination
        self.from_stage = from_stage
        self.from_image = from_image

    def emit(self):
        step = [self.kind]
        if self.from_stage is not None:
            if self.from_image is not None:
                raise RuntimeError("Ambiguous COPY source.")
            if not isinstance(self.from_stage, DockerStage):
                raise RuntimeError("Invalid from stage for COPY.")
            step.append("=".join(["--from", self.from_stage.stage_name]))
        if self.from_image is not None:
            step.append("=".join(["--from", self.from_image]))
        step.extend([self.source, self.destination])
        return " ".join(step)


class DockerRunStep:
    def __init__(self, instruction):
        self.kind = "RUN"
        self.instruction = instruction

    def emit(self):
        return " ".join([self.kind, self.instruction])


class DockerCmdStep:
    def __init__(self, command):
        self.kind = "CMD"
        self.command = command

    def emit(self):
        return " ".join([self.kind, self.command])


class DockerWorkdirStep:
    def __init__(self, work_directory):
        self.kind = "WORKDIR"
        self.work_directory = work_directory

    def emit(self):
        return " ".join([self.kind, self.work_directory])


class DockerStage:
    def __init__(self, image, stage_name):
        self.kind = "FROM"
        self.image = image
        self.stage_name = stage_name
        self.steps = []

    def add_step(self, step):
        self.steps.append(step)

    def emit(self):
        from_step = [self.kind, self.image]
        if self.stage_name is not None:
            from_step.extend(["AS", self.stage_name])
        stage = [" ".join(from_step)]
        for step in self.steps:
            stage.append(step.emit())
        return "\n".join(stage)


class DockerImage:
    def __init__(self, base_image=None):
        self.base_image = base_image
        self.docker_image_name = None
        self.stages = []
        self.command = None
        if self.base_image is not None:
            self.add_stage(self.base_image)
        self.current_stage = self.stages[-1]
        self.workspace_directory = file.Directory(local_workspace_name)
        self.os_type = None

    def copy(self, source, destination=".", from_stage=None, from_image=None):
        source_file = source
        if isinstance(source, file.File):
            self.add_file(source)
            source_file = source.name
        self.current_stage.add_step(
            DockerCopyStep(source_file, destination, from_stage, from_image))

    def run(self, instruction):
        self.current_stage.add_step(DockerRunStep(instruction))

    def cmd(self, command):
        self.current_stage.add_step(DockerCmdStep(command))

    def workdir(self, work_directory):
        self.current_stage.add_step(DockerWorkdirStep(work_directory))

    def add_stage(self, image, stage_name=None, previous_stage_name=None):
        if previous_stage_name is not None and len(self.stages) > 0:
            current_stage_name = self.current_stage.name
            if current_stage_name is None:
                self.current_stage.name = previous_stage_name
        new_stage = DockerStage(image, stage_name)
        self.stages.append(new_stage)
        self.current_stage = self.stages[-1]

    def add_file(self, host_file, sub_path=None):
        if not isinstance(host_file, file.File):
            raise RuntimeError("Invalid input file type, use File.")
        self.workspace_directory.add_file(host_file, sub_path=sub_path)

    def add_directory(self, directory):
        self.workspace_directory.add_directory(directory)

    def output_directory_structure(self, dockerfile_name="Dockerfile"):
        self.add_file(file.File(dockerfile_name, contents=self))
        self.workspace_directory.emit()

    def delete_directory_structure(self):
        self.workspace_directory.delete()

    def build(self, docker_image_name):
        self.output_directory_structure()
        self.docker_image_name = docker_image_name
        docker_build(self.workspace_directory.name, docker_image_name)
        self.delete_directory_structure()

    def push(self):
        if self.docker_image_name is None:
            raise RuntimeError(
                "Attempting to push an image that has not been built.")
        docker_push(self.docker_image_name)

    def emit(self):
        image = []
        for stage in self.stages:
            image.append(stage.emit())
        return "\n".join(image)

    def add_kubernetes(self):
        if self.os_type is None:
            raise RuntimeError(
                "Attempting to add kubectl in base docker image, "
                "consider using a derived docker image object with a "
                "specific OS type.")

    def add_kamel(self):
        if self.os_type is None:
            raise RuntimeError(
                "Attempting to add kamel executable in base docker image, "
                "consider using a derived docker image object with a "
                "specific OS type.")


# As a subclass of DockerImage this image has a particular OS type which
# determines the way things are installed and how paths are composed.
class JavaAlpineDockerImage(DockerImage):
    def __init__(self):
        DockerImage.__init__(self, base_image="adoptopenjdk/openjdk11:alpine")
        self.os_type = "linux"
        self.installed_packages = []

    def install(self, package):
        if package not in self.installed_packages:
            self.run("apk add --update " + package)
            self.installed_packages.append(package)

    def update_installed_packages(self):
        self.run("apk update && apk upgrade")

    def add_kubernetes(self):
        self.install("curl")
        self.run("curl -LO https://storage.googleapis.com/"
                 "kubernetes-release/release/`curl -s https://"
                 "storage.googleapis.com/kubernetes-release/release/"
                 "stable.txt`/bin/linux/amd64/kubectl")
        self.run("chmod +x ./kubectl")
        self.run("mv ./kubectl /usr/local/bin/")

    def add_kamel(self, version=None):
        kamel_version = v.kamel_version
        if version is not None:
            kamel_version = version
        self.copy("/usr/local/bin/kamel",
                  "/usr/local/bin/",
                  from_image="docker.io/apache/camel-k:" + kamel_version)


def add_summary_from_image(image, workspace_directory):
    # Create container from original image:
    container_id = docker_create(image)

    # Copy summary file from container to current workspace:
    docker_cp_to_host(container_id,
                      f"/{image_workspace_name}/{file.summary_file_name}", ".")

    # Remove container
    docker_rm(container_id)

    # Create a virtual summary file starting from the local file:
    summary_file = file.SummaryFile(file.summary_file_name)

    # Delete actual local file, the virtual file will continue to exist.
    summary_file.delete()

    # Add file to working directory:
    workspace_directory.add_file(summary_file)


def get_integration_directory_on_image():
    return "/".join([f"/{image_workspace_name}", built_integration_directory])


def get_routes_path_on_image(integration_file_name):
    directory = get_integration_directory_on_image()
    return "/".join([directory, "routes", integration_file_name])


def get_file_on_image(file_name):
    directory = get_integration_directory_on_image()
    return "/".join([directory, file_name])


def update_integration_file_in_image(file_name):
    copy_integration_file = ["cp"]
    copy_integration_file.append(get_file_on_image(file_name))
    copy_integration_file.append(get_routes_path_on_image(file_name))
    return " ".join(copy_integration_file)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pathlib
import platform
import subprocess
import rayvens.cli.utils as utils

summary_file_name = "summary.txt"


class FileSystemObject:
    def __init__(self, path_to_file):
        if path_to_file is None:
            raise RuntimeError("Path to file is None.")
        self.full_path = path_to_file
        if isinstance(path_to_file, str):
            self.full_path = pathlib.Path(path_to_file)
        self.name = self.full_path.name
        self.path = self.full_path.parent
        if self.path == pathlib.Path('.'):
            self.path = pathlib.Path.cwd()
            self.full_path = self.path.joinpath(self.name)

    def move(self, path):
        if path is not None and isinstance(path, str):
            path = pathlib.Path(path)
        self.path = path
        self.full_path = self.path.joinpath(self.name)

    def rename(self, name):
        self.name = name
        self.full_path = self.path.joinpath(self.name)


class File(FileSystemObject):
    def __init__(self, path, contents=None):
        FileSystemObject.__init__(self, path)
        self.contents = contents
        self.original_name = self.name
        self.original_full_path = self.full_path

    def delete(self):
        # TODO: keep track of all places where the file was emitted
        # and then when delete is invoked, remove the file from
        # all those locations.
        try:
            if self.exists_on_file_system():
                os.remove(self.full_path)
        except OSError as e:
            print("Error: %s : %s" % (self.full_path, e.strerror))

    def get_contents(self):
        if isinstance(self.contents, str):
            return self.contents
        if not hasattr(self.contents, "emit"):
            raise RuntimeError(
                "Non-string file contents must have emit method.")
        return self.contents.emit()

    def emit(self):
        if self.contents is not None:
            with open(self.full_path, mode='w') as fs_file:
                fs_file.write(self.get_contents())
        elif os.path.isfile(self.original_full_path):
            if self.original_full_path != self.full_path:
                copy_file(str(self.original_full_path), str(self.full_path))
        else:
            raise RuntimeError(f"Cannot emit file: {str(self.full_path)}")

    def exists_on_file_system(self):
        return os.path.isfile(self.full_path)

    def original_exists_on_file_system(self):
        return os.path.isfile(self.original_full_path)

    def read(self):
        file_contents = None
        if not self.original_exists_on_file_system():
            return file_contents
        with open(self.original_full_path, 'r') as f:
            file_contents = f.read()
        return file_contents


class Directory(FileSystemObject):
    def __init__(self, path):
        FileSystemObject.__init__(self, path)
        self.files = []
        self.sub_directories = []

    def move(self, path):
        FileSystemObject.move(self, path)
        for file in self.files:
            file.move(self.full_path)

    def add_file(self, file, sub_path=None, override=False):
        if not isinstance(file, File):
            raise RuntimeError("Invalid file object type provided, use File.")
        if sub_path is None and self.already_added(file) and not override:
            raise RuntimeWarning(
                f"A file with name {file.name} already added to "
                f"directory {self.full_path}")
        if sub_path is not None and not isinstance(sub_path, str):
            raise RuntimeError("Invalid sub_path type provided, use string.")
        if sub_path is not None:
            path_components = sub_path.split(os.path.sep)
            found_sub_directory = False
            for sub_directory in self.sub_directories:
                if path_components.name == path_components[0]:
                    found_sub_directory = True
                    new_sub_path = None
                    if len(path_components) > 1:
                        new_sub_path = os.path.sep.join(path_components[1:])
                    sub_directory.add_file(file, sub_path=new_sub_path)
            if not found_sub_directory:
                raise RuntimeError(
                    f"Subpath subdirectory {path_components[0]} not found.")
        else:
            file.move(self.full_path)
            self.files.append(file)

    def add_directory(self, directory):
        if not isinstance(directory, Directory):
            raise RuntimeError(
                "Invalid directory object type provided, use Directory.")
        directory.move(self.full_path)
        self.sub_directories.append(directory)

    def delete(self):
        for file in self.files:
            file.delete()
        for directory in self.sub_directories:
            directory.delete()
        try:
            self.full_path.rmdir()
        except OSError as e:
            print("Error: %s : %s" % (self.full_path, e.strerror))

    def emit(self, replace=True):
        # if replace and os.path.isdir(self.full_path):
        #     self.delete()
        os.mkdir(self.full_path)
        for file in self.files:
            file.emit()
        for directory in self.sub_directories:
            directory.emit()

    # Checks if a file has already been added to the list of files this
    # directory holds.
    def already_added(self, file):
        for existing_file in self.files:
            if existing_file.name == file.name:
                return True
        return False

    # Returns the first time it finds a file in the directory hierarchy.
    def get_file(self, file_name):
        for file in self.files:
            if file.name == file_name:
                return file
        for sub_directory in self.sub_directories:
            found_file = sub_directory.get_file_reference()
            if found_file is not None:
                return found_file
        return None


class SummaryFile(File):
    def __init__(self, path=None):
        self.kind = None
        self.launch_image = None
        self.properties = {}
        self.envvars = {}
        self._property_prefix = "property: "
        self._envvar_prefix = "envvar: "
        if path is None:
            File.__init__(self, summary_file_name, contents=None)
        else:
            File.__init__(self, path, contents=None)
            self.parse()

    def add_property(self, property_name, property_value):
        self.properties[property_name] = property_value

    def add_envvar(self, property_name, envvar):
        self.envvars[property_name] = envvar

    def get_envvars(self):
        return [key for key in self.envvars]

    def get_properties(self):
        return [key for key in self.properties]

    def parse(self):
        if not self.original_exists_on_file_system():
            raise RuntimeError(
                "Parsed summary file is not on the file system.")
        # Get kind:
        self.kind = utils._get_field_from_summary(self.original_full_path,
                                                  "kind")
        self.launch_image = utils._get_field_from_summary(
            self.original_full_path, "launch_image")
        valid_properties = utils.get_all_properties(self.kind)
        for property_name in valid_properties:
            # Property is of type:
            # property: <property_name>=<value>
            value = utils._get_field_from_summary(self.original_full_path,
                                                  property_name,
                                                  prefix=self._property_prefix)
            if value is not None:
                self.add_property(property_name, value)
            # Envvar is of type:
            # property: <property_name>=<envvar>
            envvar = utils._get_field_from_summary(self.original_full_path,
                                                   property_name,
                                                   prefix=self._envvar_prefix)
            if envvar is not None:
                self.add_envvar(property_name, envvar)

    def emit(self):
        summary = []
        summary.append(f"kind={self.kind}")
        if self.launch_image is not None:
            summary.append(f"launch_image={self.launch_image}")
        else:
            summary.append("launch_image=None")
        for prop in self.properties:
            summary.append(
                f"{self._property_prefix}{prop}={self.properties[prop]}")
        for prop in self.envvars:
            summary.append(f"{self._envvar_prefix}{prop}={self.envvars[prop]}")
        self.contents = "\n".join(summary)
        File.emit(self)


def find_executable(executable_name):
    command = ["which"]
    if platform.system() == "Windows":
        command = ["where"]
    command.append(executable_name)
    return subprocess.check_output(command).decode('utf8').strip()


def copy_file(source, destination):
    command = ["cp"]
    if platform.system() == "Windows":
        command = ["copy"]
    command.append(source)
    command.append(destination)
    subprocess.call(command)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from rayvens.core.catalog_utils import print_predefined_integrations
from rayvens.core.catalog_utils import print_requirements_summary


def rayvens_print(args):
    if args.all:
        print_predefined_integrations()
    if args.kind:
        print_requirements_summary(args.kind)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ray
import requests
import time
import threading
import os
import json
from pathlib import Path
from confluent_kafka import Consumer, Producer
from rayvens.core import kamel
from rayvens.core.mode import mode, RayvensMode
from ray.actor import ActorHandle, ActorMethod
from ray.remote_function import RemoteFunction

# Batch timeout in seconds:
batch_timeout = 10


class OutputEvent:
    def __init__(self, data, headers={}):
        self.data = data
        self.headers = headers


def get_run_mode(camel_mode, check_port, release):
    if camel_mode == 'auto' or camel_mode == 'local':
        mode.run_mode = RayvensMode.LOCAL
    elif camel_mode == 'mixed':
        mode.run_mode = RayvensMode.MIXED_OPERATOR
    elif camel_mode == 'operator':
        mode.run_mode = RayvensMode.CLUSTER_OPERATOR
    else:
        raise RuntimeError("Unsupported camel mode.")
    mode.check_port = check_port
    mode.release = release
    return mode


def wait_for_ready_integration(server_address):
    health_check_address = f"{server_address}/q/health"

    healthy_integration = False
    max_retries = 5 * 60
    while True:
        try:
            response = requests.get(health_check_address, timeout=(5, 5))
        except requests.exceptions.ConnectionError:
            time.sleep(1)
            max_retries -= 1
            if max_retries == 0:
                break
            continue
        json_response = json.loads(response.content)
        all_routes_are_up = True
        for check in json_response['checks']:
            if check['name'] == 'camel-readiness-checks' and check[
                    'status'] == 'UP':
                data = check['data']
                if data['context'] == 'UP':
                    # Ensure all routes are up.
                    route_index = 1
                    route = f'route:route{route_index}'
                    while route in data:
                        if data[route] != 'UP':
                            all_routes_are_up = False
                            break
                        route_index += 1
                        route = f'route:route{route_index}'
        if all_routes_are_up:
            healthy_integration = True
            break
        time.sleep(1)

    return healthy_integration


# Wait for an integration to reach its running state and not only that but
# also be in a state where it can immediately execute incoming requests.
def await_start(mode, integration):
    # Check logs of the integration to make sure it was installed properly.
    if not mode.is_local():
        invocation = kamel.log(mode, integration.integration_name,
                               "Installed features:")
        integration_is_running = invocation is not None
        if integration_is_running:
            print(f'Integration {integration.integration_name} is running.')
        else:
            print('Integration did not start correctly.')

    # For kafka transport the health check cannot be performed.
    if mode.transport == 'kafka':
        return True

    # Perform readiness check and wait for integration to be ready.
    server_address = mode.server_address(integration)
    healthy_integration = wait_for_ready_integration(server_address)

    if healthy_integration:
        print(f'Integration {integration.integration_name} is ready.')
        if not mode.is_local():
            return integration_is_running
        return True

    print(f'Integration {integration.integration_name} cannot be ready.')
    return False


@ray.remote(num_cpus=0)
class ProducerActor:
    def __init__(self, url):
        self.url = url

    def append(self, data):
        event_data = data
        event_headers = {}
        if isinstance(data, OutputEvent):
            event_data = data.data
            event_headers = data.headers
        try:
            if isinstance(data, Path):
                requests.post(self.url,
                              str(event_data),
                              timeout=(5, None),
                              headers=event_headers)
            else:
                requests.post(self.url,
                              event_data,
                              timeout=(5, None),
                              headers=event_headers)
        except requests.exceptions.ConnectionError:
            pass


class BatchTimeoutThread(threading.Thread):
    def __init__(self, handle):
        threading.Thread.__init__(self)
        self.stop_flag = threading.Event()
        self.received_a_message = threading.Event()
        self.handle = handle

    def run(self):
        # Wait for at least one message to be delivered:
        while not self.received_a_message:
            pass

        # Start monitoring timeout:
        while not self.stop_flag.is_set():
            self.received_a_message = threading.Event()
            time.sleep(batch_timeout)
            if not self.received_a_message.is_set():
                self.handle.flush_batch.remote()
        self.handle.flush_batch.remote()


class QueueReadThread(threading.Thread):
    def __init__(self, handle, server_address, route):
        threading.Thread.__init__(self)
        self.stop_flag = threading.Event()
        self.run_flag = threading.Event()
        self.handle = handle
        self.server_address = server_address
        self.route = route
        self.timeout_thread = BatchTimeoutThread(handle)
        self.timeout_thread.daemon = True
        self.timeout_thread.start()

    def run(self):
        while not self.stop_flag.is_set():
            try:
                response = requests.get(f'{self.server_address}{self.route}',
                                        timeout=(5, None))
                if response.status_code != 200:
                    time.sleep(1)
                    continue
                response.encoding = 'latin-1'
                self.handle.append.remote(response.text)
                self.timeout_thread.received_a_message.set()
            except requests.exceptions.ConnectionError:
                time.sleep(1)

        self.timeout_thread.stop_flag.set()
        self.run_flag.set()


# Method for sending events to the queue.
def send_to(handle, server_address, route):
    queue_reading_thread = QueueReadThread(handle, server_address, route)
    queue_reading_thread.daemon = True
    queue_reading_thread.start()
    return queue_reading_thread


# Method for sending events to the sink.
def recv_from(handle, integration_name, server_address, route):
    helper = ProducerActor.remote(f'{server_address}{route}')
    handle.send_to.remote(helper, integration_name)


@ray.remote(num_cpus=0)
class KafkaProducerActor:
    def __init__(self, name):
        self.producer = ProducerHelper(name)

    def append(self, data):
        event_data = data
        event_headers = {}
        if isinstance(data, OutputEvent):
            event_data = data.data
            event_headers = data.headers
        if isinstance(data, Path):
            self.producer.produce(str(event_data), headers=event_headers)
        else:
            self.producer.produce(event_data, headers=event_headers)


class ProducerHelper:
    def __init__(self, name):
        self.name = name
        self.producer = Producer({'bootstrap.servers': brokers()})

    def produce(self, data, headers={}):
        if data is not None:
            self.producer.produce(self.name,
                                  data.encode('utf-8'),
                                  headers=headers)


@ray.remote(num_cpus=0.05)
class KafkaConsumer(object):
    def __init__(self, kafka_transport_topic, stream_actor):
        self.stream_actor = stream_actor

        # Set up a Kafka Consumer:
        from confluent_kafka import Consumer
        self.kafka_consumer = Consumer({
            'bootstrap.servers': brokers(),
            'group.id': 'ray',
            'enable.auto.commit': True,
            'auto.offset.reset': 'earliest'
        })

        # Subscribe the Kafka Consumer to the transport topic:
        self.kafka_consumer.subscribe([kafka_transport_topic])

    def enable_consumer(self, subscribers, operator):
        self.enabled = True
        while self.enabled:
            # Retrieve message:
            msg = self.kafka_consumer.poll(1.0)
            if msg is None:
                continue
            if msg.error():
                print(f"Consumer error: {msg.error()}")
                continue

            # Decode:
            data = msg.value().decode('utf-8')

            # Check message is valid:
            if data is None:
                continue

            # Apply the operator:
            if operator is not None:
                data = ray.get(eval(operator, data))

            # Send message to subscribers:
            for name, stream_subscriber in subscribers.items():
                eval(stream_subscriber, data)
            self.stream_actor._update_timestamp.remote(time.time())

    def disable_consumer(self):
        self.enabled = False

    def close_consumer(self):
        self.kafka_consumer.close()


class ConsumerEnablingThread(threading.Thread):
    def __init__(self, handle, consumers):
        threading.Thread.__init__(self)
        self.stop_flag = threading.Event()
        self.run_flag = threading.Event()
        self.handle = handle
        self.consumers = consumers

    def run(self):
        subscribers, operator = ray.get(self.handle._fetch_processors.remote())
        try:
            consumer_references = [
                consumer.enable_consumer.remote(subscribers, operator)
                for consumer in self.consumers
            ]
            ray.get(consumer_references)
        except KeyboardInterrupt:
            for consumer in self.consumers:
                consumer.disable_consumer.remote()
        finally:
            for consumer in self.consumers:
                consumer.close_consumer.remote()

        self.run_flag.set()


class ConsumerThread(threading.Thread):
    def __init__(self, handle, consumer):
        threading.Thread.__init__(self)
        self.stop_flag = threading.Event()
        self.run_flag = threading.Event()
        self.handle = handle
        self.consumer = consumer

    def run(self):
        while not self.stop_flag.is_set():
            msg = self.consumer.poll()
            if msg.error():
                print(f'consumer error: {msg.error()}')
            else:
                self.handle.append.remote(msg.value().decode('utf-8'))

        self.run_flag.set()


def kafka_send_to(kafka_transport_topic, kafka_transport_partitions, handle):
    # If source uses a scaling transport start multiple Kafka Consumers.
    if kafka_transport_partitions > 1:
        consumers = [
            KafkaConsumer.remote(kafka_transport_topic, handle)
            for _ in range(kafka_transport_partitions)
        ]

        consumer_enabling_thread = ConsumerEnablingThread(handle, consumers)
        consumer_enabling_thread.daemon = True
        consumer_enabling_thread.start()
        return consumer_enabling_thread

    # Use kafka consumer thread to push from camel source to rayvens stream
    consumer = Consumer({
        'bootstrap.servers': brokers(),
        'group.id': 'ray',
        'auto.offset.reset': 'latest'
    })

    consumer.subscribe([kafka_transport_topic])

    consumer_thread = ConsumerThread(handle, consumer)
    consumer_thread.daemon = True
    consumer_thread.start()
    return consumer_thread


def kafka_recv_from(integration_name, kafka_transport_topic, handle):
    # use kafka producer actor to push from rayvens stream to camel sink
    helper = KafkaProducerActor.remote(kafka_transport_topic)
    handle.send_to.remote(helper, integration_name)


def brokers():
    host = os.getenv('KAFKA_SERVICE_HOST', 'localhost')
    port = os.getenv('KAFKA_SERVICE_PORT', '31093')
    return f'{host}:{port}'


def operator_brokers():
    host = os.getenv('KAFKA_SERVICE_HOST', 'kafka')
    port = os.getenv('KAFKA_SERVICE_PORT', '9092')
    return f'{host}:{port}'


def eval(f, data):
    if isinstance(f, ActorHandle):
        return f.append.remote(data)
    elif isinstance(f, ActorMethod) or isinstance(f, RemoteFunction):
        return f.remote(data)
    else:
        return f(data)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import yaml
import subprocess
import pathlib
from enum import Enum
import rayvens.core.utils as rayvens_utils
import rayvens.cli.file as file
import rayvens.cli.docker as docker
import rayvens.cli.utils as utils

all_verbs = ["get", "list", "watch", "create", "update", "patch", "delete"]
api_groups = {'jobs': ["batch", "extensions"]}
volume_base_name = "rayvens-volume"
kubernetes_tag = "kubectl"


# Enum that specifies the policy to apply when creating a new Kubernetes
# entity when a Kubernetes entity with the same name already exists. The
# policy can only be applied at the time the entity is created to ensure the
# uniqueness of the name.
class CreatePolicy(Enum):
    # Create a new Kubernetes entity with a new unique name:
    NEW = 1

    # Skip creation of a new entity and reuse the one that exists already:
    REUSE = 2

    # Delete the old entity and create a new one with the same name:
    REPLACE = 3


def get_deployment(name, namespace, registry, args, with_job_launcher,
                   integration_config_map, k8s_client):
    # Kubernetes deployment options:
    replicas = 1
    integration_name = utils.get_kubernetes_integration_name(name)
    entrypoint_name = utils.get_kubernetes_entrypoint_name(name)
    label_name = utils.get_kubernetes_label_name(name)
    image_name = utils.extract_image_name(args)
    container = Container(image_name, args.image, image_pull_policy="Always")

    # Update integration file from host file.
    container.update_integration(integration_config_map)
    if with_job_launcher:
        # Service account for the job lunch permissions:
        service_account = ServiceAccount(utils.job_launcher_service_account,
                                         namespace)

        # Create the cluster role specifying the resource type and the
        # allowable actions (verbs). By default all verbs are enabled if
        # none are specified.
        cluster_role = ClusterRole(utils.job_manager_role, namespace)
        cluster_role.add_rule("jobs")

        # Create a cluster role binding between the service account and
        # the cluster role defined above.
        cluster_role_binding = ClusterRoleBinding(
            utils.job_launcher_cluster_role_binding, [service_account],
            cluster_role)
        cluster_role_binding.create(k8s_client)

    # Assemble the pod that the deployment will be managing.
    managed_pod = Pod("managed_pod", namespace)

    # Ensure the pod has the ability to launch jobs by addint the service
    # account created above.
    if with_job_launcher:
        managed_pod.add_service_account(service_account)
    managed_pod.add_container(container)
    managed_pod.add_label("integration", label_name)

    # Create the deployment.
    deployment = Deployment(integration_name,
                            managed_pod,
                            namespace=namespace,
                            replicas=replicas)

    # Create a NodePort service to enable outside communication.
    node_port_spec = NodePortSpec()
    node_port_spec.add_selector("integration", label_name)

    # Fetch unused port:
    node_port_spec.add_port(3000,
                            rayvens_utils.random_port(start=30000, end=32767))
    service = Service(entrypoint_name, node_port_spec, namespace=namespace)

    # Add the service to the deployment.
    deployment.add_service(service)
    return deployment


def kubectl_create_configmap(name, namespace, embedded_file_path):
    command = ["kubectl", "create", "configmap", name]

    # Set the namespace:
    if namespace is not None:
        command.append("-n")
        command.append(namespace)

    # From file:
    command.append("--from-file")
    command.append(embedded_file_path)

    # Wait for kubectl command to finish before returning:
    outcome = subprocess.run(command)

    if outcome.returncode == 0:
        utils.PRINT(f"configmap object {name} created successfully.",
                    tag=kubernetes_tag)
    else:
        utils.PRINT(f"configmap object {name} creation failed.",
                    tag=kubernetes_tag)


def kubectl_get_yaml(entity_type, name, namespace):
    command = ["kubectl", "get", entity_type, name]

    # Namespace.
    if namespace is not None:
        command.append("-n")
        command.append(namespace)

    # From file:
    command.append("-o")
    command.append("yaml")

    # Wait for kubectl command to finish before returning:
    outcome = subprocess.run(command)
    if outcome.returncode == 0:
        utils.PRINT("Kubernetes get ended successfully.", tag=kubernetes_tag)
    else:
        utils.PRINT("Kubernetes get failed.", tag=kubernetes_tag)


class KubeEntity:
    def __init__(self, name, namespace=None, application=None):
        # Name of the Kubernetes entity, the name will be stored
        # as a piece of metadata information.
        self.name = name

        # The namespace in which the kubernetes entity will be launched.
        self.namespace = namespace
        self._metadata = {}
        self._metadata["name"] = name
        if namespace is not None:
            self._metadata["namespace"] = namespace
        self._labels = {}
        self._count = 0
        if application is not None:
            self._labels["app"] = application
        self.template = None

    def add_label(self, name, value):
        self._labels[name] = value

    def add_metadata(self, entry_name, entry_value):
        self._metadata[entry_name] = entry_value

    def app(self, application_name):
        self._labels["app"] = application_name

    def configuration(self):
        return yaml.dump(self._specification())

    def file(self):
        return file.File(self.name + ".yaml", contents=self.configuration())

    def create(self, k8s_client):
        import kubernetes.utils as kube_utils

        # Kubernetes entities may or may not support specifications. For those
        # without a specification method, create the kubernetes entity starting
        # from the configuration.
        if hasattr(self, "_specification"):
            try:
                kube_utils.create_from_dict(k8s_client, self._specification())
            except kube_utils.FailToCreateError as creation_error:
                utils.PRINT(
                    f"Failed to create kubernetes entity {creation_error}",
                    tag=kubernetes_tag)
        else:
            workspace_directory = file.Directory("kube-workspace")
            file_name = utils.get_kubernetes_deployment_file_name(self.name)
            spec_file = file.File(file_name, contents=self.configuration())

            # Materialize file:
            workspace_directory.add_file(spec_file)
            workspace_directory.emit()

            # Create entity:
            try:
                kube_utils.create_from_yaml(k8s_client,
                                            str(spec_file.full_path))
            except kube_utils.FailToCreateError as creation_error:
                utils.PRINT(
                    f"Failed to create kubernetes entity {creation_error}.",
                    tag=kubernetes_tag)
            else:
                utils.PRINT(f"{self.name} successfully created.",
                            tag=kubernetes_tag)
            workspace_directory.delete()

    def _delete_namespaced(self, api_method, name):
        namespace = self.namespace
        if self.namespace is None:
            namespace = "default"
        from kubernetes.client.rest import ApiException
        try:
            api_method(name, namespace)
        except ApiException as deletion_error:
            utils.PRINT(f"Failed to delete kubernetes entity {deletion_error}",
                        tag=kubernetes_tag)
        else:
            utils.PRINT(f"Kubernetes entity {self.name} deleted successfully",
                        tag=kubernetes_tag)

    def _delete(self, api_method, name):
        from kubernetes.client.rest import ApiException
        try:
            api_method(name)
        except ApiException as deletion_error:
            utils.PRINT(f"Failed to delete kubernetes entity {deletion_error}",
                        tag=kubernetes_tag)
        else:
            utils.PRINT(f"Kubernetes entity {self.name} deleted successfully",
                        tag=kubernetes_tag)

    def _exists_namespaced(self, api_method):
        namespace = self.namespace
        if self.namespace is None:
            namespace = "default"
        from kubernetes.client.rest import ApiException
        try:
            api_response = api_method(namespace)
        except ApiException as check_error:
            utils.PRINT(
                f"Exception when fetching kubernetes entities {check_error}",
                tag=kubernetes_tag)
        return api_response

    def _exists(self, api_method):
        from kubernetes.client.rest import ApiException
        try:
            api_response = api_method()
        except ApiException as check_error:
            utils.PRINT(
                f"Exception when fetching kubernetes entities {check_error}",
                tag=kubernetes_tag)
        return api_response


# apiVersion: v1
# kind: ConfigMap
# metadata:
#   name: game-config-env-file
#   namespace: default
# data:
#   filename.ext: |+
# <file_contents>


class ConfigMap(KubeEntity):
    def __init__(self, embedded_file=None, namespace=None):
        KubeEntity.__init__(self, "config-map", namespace=namespace)
        self.api_version = "v1"
        self.kind = "ConfigMap"
        if embedded_file is not None and not isinstance(
                embedded_file, file.File):
            raise RuntimeError("Input file must be of File type.")
        self.embedded_file = embedded_file
        self._create_policy = CreatePolicy.NEW

    def _specification(self):
        raise RuntimeError("Not implemented yet for this kubernetes entity.")

    def create(self, k8s_client, create_policy=None):
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.NEW:
            while True:
                self.name = volume_base_name + "-" + str(self._count)
                if not self.exists(k8s_client):
                    self.embedded_file.emit()
                    utils.PRINT(
                        f"Create {self.name} in namespace {self.namespace}.",
                        tag=kubernetes_tag)
                    kubectl_create_configmap(self.name, self.namespace,
                                             self.embedded_file.full_path)
                    # kubectl_get_yaml("configmap", self.name, self.namespace)
                    self.embedded_file.delete()
                    self._count += 1
                    break
                self._count += 1
        else:
            raise RuntimeError(
                f"Create policy {create_policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        config_map_list = KubeEntity._exists_namespaced(
            self, api_instance.list_namespaced_config_map)
        for item in config_map_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        if starting_with is not None:
            config_map_list = KubeEntity._exists_namespaced(
                self, api_instance.list_namespaced_config_map)
            for item in config_map_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete_namespaced(
                        self, api_instance.delete_namespaced_config_map,
                        item.metadata.name)
        else:
            KubeEntity._delete_namespaced(
                self, api_instance.delete_namespaced_config_map, self.name)


#  apiVersion: v1
#  kind: Pod
#  metadata:
#    name: rss-site
#    labels:
#      app: web
#  spec:
#    containers:
#      - name: front-end
#        image: nginx
#        ports:
#          - containerPort: 80
#      - name: rss-reader
#        image: nickchase/rss-php-nginx:v1
#        ports:
#          - containerPort: 88

#   containers:
#   - image: k8s.gcr.io/test-webserver
#     name: test-container
#     volumeMounts:
#     - mountPath: /test-pd
#       name: test-volume
#   volumes:
#   - name: test-volume
#     hostPath:
#       # directory location on host
#       path: /data


class Volume:
    def __init__(self, name, namespace=None):
        self.name = name
        self.namespace = namespace
        self.kind = None
        self.host_file_path = None
        self.mount_file_path = None

    def _update_integration_file(self, integration_config_map):
        self.kind = "configMap"
        self.host_file_path = integration_config_map.embedded_file.full_path
        self.mount_file_path = docker.get_file_on_image(
            self.host_file_path.name)
        if isinstance(self.mount_file_path, str):
            self.mount_file_path = pathlib.Path(self.mount_file_path)

    def _volume_specification(self):
        spec = {'name': self.name}
        volume_spec = {}
        if self.kind == "hostPath":
            volume_spec['path'] = str(self.host_file_path)
        if self.kind == "configMap":
            volume_spec['name'] = self.name
            item_spec = {
                'key': str(self.host_file_path.name),
                'path': str(self.host_file_path.name)
            }
            volume_spec['items'] = [item_spec]
        spec[self.kind] = volume_spec
        return spec

    def _volume_mount_specification(self):
        spec = {'name': self.name}
        if self.kind == "hostPath" or self.kind == "configMap":
            spec['mountPath'] = str(self.mount_file_path)
            if self.mount_file_path is not None:
                spec['subPath'] = str(self.mount_file_path.name)
        return spec

    def delete(self):
        self.config_map.delete()


class Container:
    def __init__(self, name, image, image_pull_policy=None):
        self.name = name
        self.image = image
        self.image_pull_policy = image_pull_policy
        self.command = None
        self._envvars = []
        self._ports = []
        self._volumes = []
        self._volume_mounted_count = 0

    def add_port(self, port_name, value):
        self._ports.append({port_name: value})

    def add_envvar(self, env_var_name, value):
        self._envvars.append({"name": env_var_name, "value": value})

    def update_integration(self, integration_config_map):
        volume = Volume(integration_config_map.name)
        volume._update_integration_file(integration_config_map)
        self._volumes.append(volume)
        self._volume_mounted_count += 1

    def _specification(self):
        spec = {'name': self.name, 'image': self.image}

        if len(self._ports) > 0:
            spec['ports'] = self._ports

        if len(self._envvars) > 0:
            spec['env'] = self._envvars

        if self.command is not None:
            spec['command'] = self.command

        if self.image_pull_policy is not None:
            spec['imagePullPolicy'] = self.image_pull_policy

        if len(self._volumes) > 0:
            volume_mounts_spec = []
            for volume in self._volumes:
                volume_mounts_spec.append(volume._volume_mount_specification())
            spec['volumeMounts'] = volume_mounts_spec

        return spec

    def _volumes_specifications(self):
        volume_specs = []
        for volume in self._volumes:
            volume_specs.append(volume._volume_specification())
        return volume_specs


class Pod(KubeEntity):
    def __init__(self, name, namespace=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "v1"
        self.kind = "Pod"
        self.containers = []
        self._service_account = None
        self._create_policy = CreatePolicy.NEW

    def add_container(self, container):
        self.containers.append(container)

    def add_service_account(self, service_account):
        self._service_account = service_account
        # TODO: when this Pod is launched, make sure that
        # the service account and the roles/bindings using
        # that service account are up and running as well.

    def _pod_specification(self):
        metadata = {'name': self.name}
        if self.namespace is not None:
            metadata['namespace'] = self.namespace

        if len(self._labels) > 0:
            metadata['labels'] = self._labels

        container_specs = []
        volume_specs = []
        for container in self.containers:
            if not isinstance(container, Container):
                raise RuntimeError("Invalid container used in pod.")
            container_specs.append(container._specification())
            volume_specs.extend(container._volumes_specifications())

        pod_spec = {}
        if self._service_account is not None:
            if not isinstance(self._service_account, ServiceAccount):
                raise RuntimeError(
                    "Invalid service account type, use ServiceAccount.")
            pod_spec['serviceAccountName'] = self._service_account.name
        pod_spec['containers'] = container_specs

        if len(volume_specs) > 0:
            pod_spec['volumes'] = volume_specs

        spec = {'metadata': metadata, 'spec': pod_spec}
        return spec

    def _specification(self):
        spec = {'apiVersion': self.api_version, 'kind': self.kind}
        spec.update(self._pod_specification())
        return spec


#           "apiVersion: batch/v1" + newLine
#         + "kind: Job" + newLine
#         + "metadata:" + newLine
#         + "  name: "+ jobName + newLine
#         + "spec:" + newLine
#         + "  template:" + newLine
#         + "    spec:" + newLine
#         + "      containers:" + newLine
#         + "      - name: " + jobName + "-container" + newLine
#         + "        image: adoptopenjdk/openjdk11:alpine" + newLine
#         + "        env:" + newLine
#         + "        - name: TEST_EVENT" + newLine
#         + "          value: \\\"" + eventContents + "\\\"" + newLine
#         + "        command: {job_command}" + newLine
#         + "      restartPolicy: Never" + newLine
#         + "  backoffLimit: 4" + newLine;


class Job(KubeEntity):
    def __init__(self, name, namespace=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "batch/v1"
        self.kind = "Job"
        self.containers = []
        self.restart_policy = "Never"
        self.backoff_limit = 4
        self._create_policy = CreatePolicy.NEW

    def add_container(self, container):
        self.containers.append(container)

    def _specification(self):
        metadata = {'name': self.name}
        if self.namespace is not None:
            metadata['namespace'] = self.namespace

        container_specs = []
        for container in self.containers:
            if not isinstance(container, Container):
                raise RuntimeError("Invalid container used in pod.")
            container_specs.append(container._specification())

        spec = {
            'apiVersion': self.api_version,
            'kind': self.kind,
            'metadata': metadata,
            'spec': {
                'template': {
                    'spec': {
                        'containers': container_specs
                    },
                    'restartPolicy': self.restart_policy
                },
                'backoffLimit': self.backoff_limit
            }
        }
        return spec


# apiVersion: v1
# kind: ServiceAccount
# metadata:
#   name: {job_launcher_service_account}
#   namespace: {namespace}


class ServiceAccount(KubeEntity):
    def __init__(self, name, namespace=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "v1"
        self.kind = "ServiceAccount"
        self._create_policy = CreatePolicy.REUSE

    def _specification(self):
        metadata = {'name': self.name}
        if self.namespace is not None:
            metadata['namespace'] = self.namespace
        spec = {
            'apiVersion': self.api_version,
            'kind': self.kind,
            'metadata': metadata
        }
        return spec

    def create(self, k8s_client, create_policy=None):
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.REUSE:
            if not self.exists(k8s_client):
                KubeEntity.create(self, k8s_client)
        else:
            raise RuntimeError(f"Policy {policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        service_account_list = KubeEntity._exists_namespaced(
            self, api_instance.list_namespaced_service_account)
        for item in service_account_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        if starting_with is not None:
            service_account_list = KubeEntity._exists_namespaced(
                self, api_instance.list_namespaced_service_account)
            for item in service_account_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete_namespaced(
                        self, api_instance.delete_namespaced_service_account,
                        item.metadata.name)
        elif self.exists(k8s_client):
            KubeEntity._delete_namespaced(
                self, api_instance.delete_namespaced_service_account,
                self.name)


# apiVersion: rbac.authorization.k8s.io/v1
# kind: ClusterRole
# metadata:
#   namespace: {namespace}
#   name: {job_manager_role}
# rules:
# - apiGroups: ["batch", "extensions"]
#   resources: ["jobs", "jobs/status", "jobs/exec", "jobs/log"]
#   verbs: ["get", "list", "watch", "create", "update", "patch", "delete"]


class ClusterRoleRule:
    def __init__(self, resources, verbs):
        self.resources = resources
        self.verbs = verbs
        if len(verbs) == 0:
            self.verbs = all_verbs
        else:
            self._check_verbs()
        self.api_groups = []
        self._update_api_group()

    def add_resource(self, resource):
        self.resources.append(resource)
        self._update_api_group()

    def add_verb(self, verb):
        if verb is all_verbs and verb not in self.verbs:
            self.verbs.append(verb)

    def _update_api_group(self):
        for resource in self.resources:
            for api_group in api_groups[resource]:
                if api_group not in self.api_groups:
                    self.api_groups.append(api_group)

    def _check_verbs(self):
        for verb in self.verbs:
            if verb not in all_verbs:
                raise RuntimeError(f"Invalid verb {verb}.")


class ClusterRole(KubeEntity):
    def __init__(self, name, namespace=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "rbac.authorization.k8s.io/v1"
        self.kind = "ClusterRole"
        self.rules = []
        self._create_policy = CreatePolicy.REUSE

    def add_rule(self, resources, verbs=[]):
        resources_list = resources
        if not isinstance(resources, list):
            resources_list = [resources]
        self.rules.append(ClusterRoleRule(resources_list, verbs))

    def create(self, k8s_client, create_policy=None):
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.REUSE:
            if not self.exists(k8s_client):
                KubeEntity.create(self, k8s_client)
        else:
            raise RuntimeError(f"Policy {policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.RbacAuthorizationV1Api(k8s_client)
        cluster_role_list = KubeEntity._exists(self,
                                               api_instance.list_cluster_role)
        for item in cluster_role_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.RbacAuthorizationV1Api(k8s_client)
        if starting_with is not None:
            role_binding_list = KubeEntity._exists(
                self, api_instance.list_cluster_role)
            for item in role_binding_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete(self, api_instance.delete_cluster_role,
                                       item.metadata.name)
        elif self.exists(k8s_client):
            KubeEntity._delete(self, api_instance.delete_cluster_role,
                               self.name)

    # ISSUE: ClusterRoleRule cannot be encoded using the specification
    # notation. This is due to the special syntax for rule which does
    # not translate using the Python yaml package.
    # def _specification(self):
    #     metadata = {'name': self.name}
    #     if self.namespace is not None:
    #         metadata['namespace'] = self.namespace
    #     rules = []
    #     for rule in self.rules:
    #         rules.append(rule._specification())
    #     spec = {
    #         'apiVersion': self.api_version,
    #         'kind': self.kind,
    #         'metadata': metadata,
    #         'rules': rules
    #     }
    #     return spec

    def configuration(self):
        result = [" ".join(["apiVersion:", self.api_version])]
        result.append(" ".join(["kind:", self.kind]))
        result.append("metadata:")

        indent = "  "
        result.append(indent + " ".join(["name:", self.name]))
        result.append(indent + " ".join(["namespace:", self.namespace]))
        result.append("rules:")
        for rule in self.rules:
            result.append("- " + " ".join(
                ["apiGroups:",
                 output_list_of_strings(rule.api_groups)]))
            result.append(indent + " ".join(
                ["resources:",
                 output_list_of_strings(rule.resources)]))
            result.append(indent + " ".join(
                ["verbs:", output_list_of_strings(rule.verbs)]))
        return "\n".join(result)

    def file(self):
        return file.File(self.name + ".yaml",
                         contents=self.configuration(self))


# apiVersion: rbac.authorization.k8s.io/v1
# kind: ClusterRoleBinding
# metadata:
#   name: {job_launcher_cluster_role_binding}
# subjects:
# - kind: ServiceAccount
#   name: {job_launcher_service_account}
#   namespace: {namespace}
# roleRef:
#   kind: ClusterRole
#   name: {job_manager_role}
#   apiGroup: rbac.authorization.k8s.io


class ClusterRoleBinding(KubeEntity):
    def __init__(self, name, subjects, cluster_role):
        KubeEntity.__init__(self, name, namespace=None)
        self.api_version = "rbac.authorization.k8s.io/v1"
        self.kind = "ClusterRoleBinding"
        self.subjects = subjects
        self.cluster_role = cluster_role
        self._create_policy = CreatePolicy.REUSE
        if cluster_role is not None and not isinstance(cluster_role,
                                                       ClusterRole):
            raise RuntimeError("Input role must be of ClusterRole type.")

    def create(self, k8s_client, create_policy=None):
        # Create all subjects:
        for subject in self.subjects:
            if not hasattr(subject, "create"):
                raise RuntimeError(
                    "Cluster role binding subject has no create method.")
            subject.create(k8s_client)

        # Create cluster role:
        if not hasattr(self.cluster_role, "create"):
            raise RuntimeError(
                "Cluster role binding subject has no create method.")
        self.cluster_role.create(k8s_client)

        # Create cluster role binding:
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.REUSE:
            if not self.exists(k8s_client):
                KubeEntity.create(self, k8s_client)
        else:
            raise RuntimeError(f"Policy {policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.RbacAuthorizationV1Api(k8s_client)
        cluster_role_binding_list = KubeEntity._exists(
            self, api_instance.list_cluster_role_binding)
        for item in cluster_role_binding_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.RbacAuthorizationV1Api(k8s_client)
        if starting_with is not None:
            cluster_role_binding_list = KubeEntity._exists(
                self, api_instance.list_cluster_role_binding)
            for item in cluster_role_binding_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete(
                        self, api_instance.delete_cluster_role_binding,
                        item.metadata.name)
        elif self.exists(k8s_client):
            KubeEntity._delete(self, api_instance.delete_cluster_role_binding,
                               self.name)

    def _specification(self):
        metadata = {'name': self.name}
        subjects = []
        for subject in self.subjects:
            subject_spec = {}
            subject_spec['kind'] = subject.kind
            subject_spec['name'] = subject.name
            subject_spec['namespace'] = subject.namespace
            subjects.append(subject_spec)
        cluster_role = {}
        cluster_role['kind'] = self.cluster_role.kind
        cluster_role['name'] = self.cluster_role.name
        cluster_role['apiGroup'] = 'rbac.authorization.k8s.io'
        spec = {
            'apiVersion': self.api_version,
            'kind': self.kind,
            'metadata': metadata,
            'subjects': subjects,
            'roleRef': cluster_role
        }
        return spec


# apiVersion: v1
# kind: Service
# metadata:
#   name: {entrypoint_name}
#   namespace: {namespace}
# spec:
#   ttype: NodePort
#   selector:
#     integration: {label_name}
#   ports:
#   - port: 3000
#     targetPort: 3000
#     nodePort: 30001


class ServiceSpec:
    def __init__(self):
        pass


class NodePortSpec(ServiceSpec):
    def __init__(self):
        ServiceSpec.__init__(self)
        self.type = "NodePort"
        self.ports = []
        self.selectors = {}

    def add_selector(self, selector_name, label_name):
        self.selectors[selector_name] = label_name

    def add_port(self, port, node_port, target_port=None):
        ports = {}
        ports['port'] = port
        ports['nodePort'] = node_port
        if target_port is not None:
            ports['targetPort'] = target_port
        else:
            ports['targetPort'] = port
        self.ports.append(ports)

    def _specification(self):
        spec = {
            'type': self.type,
        }
        if len(self.selectors) > 0:
            spec['selector'] = self.selectors
        if len(self.ports) == 0:
            raise RuntimeError("NodePort service does not expose any ports.")
        spec['ports'] = self.ports
        return spec


class Service(KubeEntity):
    def __init__(self, name, spec=None, namespace=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "v1"
        self.kind = "Service"
        self.spec = spec
        self._create_policy = CreatePolicy.NEW

    def create(self, k8s_client, create_policy=None):
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.NEW:
            old_name = self.name
            while True:
                if not self.exists(k8s_client):
                    KubeEntity.create(self, k8s_client)
                    break
                self.name = old_name + "-" + str(self._count)
                self._count += 1
        else:
            raise RuntimeError(f"Policy {policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        service_list = KubeEntity._exists_namespaced(
            self, api_instance.list_namespaced_service)
        for item in service_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.CoreV1Api(k8s_client)
        if starting_with is not None:
            service_list = KubeEntity._exists_namespaced(
                self, api_instance.list_namespaced_service)
            for item in service_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete_namespaced(
                        self, api_instance.delete_namespaced_service,
                        item.metadata.name)
        elif self.exists(k8s_client):
            KubeEntity._delete_namespaced(
                self, api_instance.delete_namespaced_service, self.name)

    def _specification(self):
        metadata = {'name': self.name}
        if self.namespace is not None:
            metadata['namespace'] = self.namespace

        service_spec = {}
        if isinstance(self.spec, dict):
            service_spec = self.spec
        elif isinstance(self.spec, ServiceSpec):
            service_spec = self.spec._specification()
        else:
            raise RuntimeError(
                "Invalid service specification type. Use dict or ServiceSpec")
        spec = {
            'apiVersion': self.api_version,
            'kind': self.kind,
            'metadata': metadata,
            'spec': service_spec
        }
        return spec


# apiVersion: apps/v1
# kind: Deployment
# metadata:
#   name: {integration_name}
#   namespace: {namespace}
# spec:
#   replicas: {replicas}
#   selector:
#     matchLabels:
#       integration: {label_name}
#   template:
#     metadata:
#       labels:
#         integration: {label_name}
#     spec: {container}


class Deployment(KubeEntity):
    def __init__(self, name, managed_pod=None, namespace=None, replicas=None):
        KubeEntity.__init__(self, name, namespace=namespace)
        self.api_version = "apps/v1"
        self.kind = "Deployment"
        self.replicas = replicas
        self.match_labels = {}
        self.managed_pod = managed_pod
        if managed_pod is not None:
            self.match_labels.update(managed_pod._labels)
        self._services = []
        self._create_policy = CreatePolicy.NEW

    def add_match_label(self, selector_name, label_name):
        self.match_labels[selector_name] = label_name

    def add_service(self, service):
        self._services.append(service)

    def add_managed_pod(self, managed_pod):
        self.managed_pod = managed_pod
        if managed_pod is not None:
            self.match_labels.update(managed_pod._labels)

    def create(self, k8s_client, create_policy=None):
        # Create all services:
        for service in self._services:
            if not hasattr(service, "create"):
                raise RuntimeError("Service does not have create method.")
            service.create(k8s_client)

        # Create deployment:
        policy = self._create_policy
        if create_policy is not None and isinstance(create_policy,
                                                    CreatePolicy):
            policy = create_policy
        if policy == CreatePolicy.NEW:
            old_name = self.name
            while True:
                if not self.exists(k8s_client):
                    KubeEntity.create(self, k8s_client)
                    break
                self.name = old_name + "-" + str(self._count)
                self._count += 1
        else:
            raise RuntimeError(f"Policy {policy} not implemented.")

    def exists(self, k8s_client):
        from kubernetes import client
        api_instance = client.AppsV1Api(k8s_client)
        deployment_list = KubeEntity._exists_namespaced(
            self, api_instance.list_namespaced_deployment)
        for item in deployment_list.items:
            if item.metadata.name == self.name:
                return True
        return False

    def delete(self, k8s_client, starting_with=None):
        from kubernetes import client
        api_instance = client.AppsV1Api(k8s_client)
        if starting_with is not None:
            deployment_list = KubeEntity._exists_namespaced(
                self, api_instance.list_namespaced_deployment)
            for item in deployment_list.items:
                if item.metadata.name.startswith(starting_with):
                    KubeEntity._delete_namespaced(
                        self, api_instance.delete_namespaced_deployment,
                        item.metadata.name)
        elif self.exists(k8s_client):
            KubeEntity._delete_namespaced(
                self, api_instance.delete_namespaced_deployment, self.name)

    def _specification(self):
        metadata = {'name': self.name}
        if self.namespace is not None:
            metadata['namespace'] = self.namespace

        deployment_spec = {}

        # Replicas
        replicas = 1
        if self.replicas is not None:
            replicas = self.replicas
        deployment_spec['replicas'] = replicas

        # Select pods
        pod_selector = {"matchLabels": self.match_labels}
        if len(self.match_labels) > 0:
            deployment_spec['selector'] = pod_selector

        # Pod being deployed
        if not isinstance(self.managed_pod, Pod):
            raise RuntimeError("Deployment pod is not of type Pod.")
        deployment_spec['template'] = self.managed_pod._pod_specification()

        spec = {
            'apiVersion': self.api_version,
            'kind': self.kind,
            'metadata': metadata,
            'spec': deployment_spec
        }
        return spec

    def configuration(self):
        final_config = [yaml.dump(self._specification())]
        for deployment_service in self._services:
            final_config.append(deployment_service.configuration())
        return "\n---\n".join(final_config)


def output_list_of_strings(list_of_strings):
    list_components = []
    for entry in list_of_strings:
        list_components.append("'" + entry + "'")

    return "[" + ",".join(list_components) + "]"



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import psutil
import signal
import subprocess
import sys

is_local = sys.argv[1] == "kamel"

# Start child process.
process = None
if is_local:
    process = subprocess.Popen(sys.argv[1:], start_new_session=True)
else:
    process = subprocess.Popen(sys.argv[3:], start_new_session=True)


def clean_up():
    if sys.platform == "win32":
        os.kill(process.pid, signal.CTRL_C_EVENT)
    else:
        os.killpg(os.getpgid(process.pid), signal.SIGKILL)

    # Delete integration file.
    os.remove(sys.argv[-1])


def sigterm_handler(*args):
    clean_up()
    sys.exit()


signal.signal(signal.SIGTERM, sigterm_handler)

# Wait for parent process.
psutil.wait_procs([psutil.Process().parent()])

# Terminate child process.
if is_local:
    clean_up()
else:
    # Operator commands can only be terminated by running a kamel delete
    # command.
    command = ["kamel", "delete", sys.argv[1], "-n", sys.argv[2]]
    subprocess.Popen(command, start_new_session=True)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from rayvens.cli import utils
import rayvens.cli.kubernetes as kube
import rayvens.cli.docker as docker


def delete(args):
    # Set verbosity:
    utils.verbose = args.verbose

    if args.deployed:
        if args.name is not None:
            delete_deployment(args)
    else:
        if args.name is not None:
            delete_local_deployment(args)

    if args.all_jobs is not None:
        delete_all_jobs(args)


def delete_all_jobs(args):
    namespace = "default"
    if args.namespace is not None:
        namespace = args.namespace

    prefix = args.all_jobs

    # Delete the jobs that start with the provided string.
    from kubernetes import client, config
    config.load_kube_config()
    k8s_client = client.BatchV1Api(client.ApiClient())

    exception_occurred = False
    try:
        api_response = k8s_client.list_namespaced_job(namespace)
        body = client.V1DeleteOptions(propagation_policy='Background')
        for job in api_response.items:
            if job.metadata.name.startswith(prefix):
                k8s_client.delete_namespaced_job(job.metadata.name,
                                                 namespace,
                                                 body=body)
    except client.exceptions.ApiException:
        exception_occurred = True

    # Delete the jobs that start with the provided string.
    k8s_client = client.CoreV1Api(client.ApiClient())

    if not exception_occurred:
        print(f"Successfully deleted jobs starting with {prefix} from "
              f"{namespace} namespace.")


def delete_deployment(args, with_job_launcher_privileges=True):
    if args.name is None:
        raise RuntimeError("No integration name provided.")

    namespace = "default"
    if args.namespace is not None:
        namespace = args.namespace

    # Delete kubernetes deployment for integration.
    from kubernetes import client, config
    config.load_kube_config()
    k8s_client = client.ApiClient()

    # Delete the integration deployment.
    deployment_name = utils.get_kubernetes_integration_name(args.name)
    deployment = kube.Deployment(deployment_name, namespace=namespace)
    deployment.delete(k8s_client, starting_with=deployment_name)

    # Delete the entrypoint service.
    entrypoint_service_name = utils.get_kubernetes_entrypoint_name(args.name)
    service = kube.Service(entrypoint_service_name, namespace=namespace)
    service.delete(k8s_client, starting_with=entrypoint_service_name)

    # Delete configMap for updating the integration file.
    config_map = kube.ConfigMap(namespace=namespace)
    config_map.delete(k8s_client, kube.volume_base_name)

    if with_job_launcher_privileges:
        # Delete service account:
        # job-launcher-service-account
        name = utils.job_launcher_service_account
        service_account = kube.ServiceAccount(name, namespace=namespace)
        service_account.delete(k8s_client)

        # Delete cluster role binding:
        # job-launcher-service-account
        name = utils.job_launcher_cluster_role_binding
        cluster_role_binding = kube.ClusterRoleBinding(name, [], None)
        cluster_role_binding.delete(k8s_client)

        # Delete cluster role:
        # job-manager-role
        name = utils.job_manager_role
        cluster_role_binding = kube.ClusterRole(name, namespace=namespace)
        cluster_role_binding.delete(k8s_client)


def delete_local_deployment(args):
    if args.name is None:
        raise RuntimeError("No integration name provided.")

    containers = docker.docker_container_ls()

    for container_name in containers:
        if container_name.startswith(args.name):
            docker.docker_kill(containers[container_name])
            docker.docker_rm(containers[container_name])



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
from rayvens.core.common import brokers
from rayvens.core.utils import random_port, create_partitioned_topic
from rayvens.core.name import name_integration
from rayvens.core import catalog
from rayvens.core import kamel
from rayvens.core import kubernetes


class Integration:
    def __init__(self, stream_name, source_sink_name, config):
        self.stream_name = stream_name
        self.source_sink_name = source_sink_name

        # Integration must be marked as either a source or sink.
        if 'integration_type' not in config:
            raise RuntimeError(
                "Unreachable: integration must have an integration_type.")
        self.config = config

        # Sinks may accept only certain message types ensure that only
        # messages with the correct type reach a particular sink.
        if self.config['integration_type'] == 'sink':
            self.input_restrictions = catalog.input_restriction(config)

        self.integration_name = name_integration(self.stream_name,
                                                 self.source_sink_name)

        # Establish kafka transport topic name:
        self.kafka_transport_topic = self.integration_name
        if "kafka_transport_topic" in config:
            self.kafka_transport_topic = config["kafka_transport_topic"]

        # Establish kafka transport topic partitions:
        self.kafka_transport_partitions = 1
        if "kafka_transport_partitions" in config and \
           config["kafka_transport_partitions"] > 1:
            self.kafka_transport_partitions = config[
                "kafka_transport_partitions"]

        self.port = None
        self.invocation = None
        self.service_name = None
        self.server_address = None
        self.environment_preparators = []
        self.thread = None

    def invoke_local_run(self, mode, integration_content):
        self.port = random_port(check_port=mode.check_port)
        self.invocation = kamel.local_run(
            [integration_content],
            mode,
            self.integration_name,
            port=self.port,
            integration_type=self.config['integration_type'])

    def invoke_run(self, mode, integration_content):
        self.invocation = kamel.run(
            [integration_content],
            mode,
            self.integration_name,
            integration_type=self.config['integration_type'])

        # If running in mixed mode, i.e. Ray locally and kamel in the cluster,
        # then we have to also start a service the allows outside processes to
        # send data to the sink.
        if mode.is_mixed():
            self.service_name = "-".join(["service", self.integration_name])
            kubernetes.create_kamel_external_service(mode, self.service_name,
                                                     self.integration_name)

    # Function which disconnects an integration whether it is an integration
    # created using the kamel operator and `kamel run` or an integration
    # created using `kamel local run`.
    def disconnect(self, mode):
        # Set stop_flag to true.
        if self.thread is not None:
            # Signal the end of ongoing loops.
            self.thread.stop_flag.set()

            # Wait for threads to perform perform/finish the current requests.
            # Small 10 s timer to wait for request completion.
            countdown = 10
            while not self.thread.run_flag.is_set():
                time.sleep(1)
                if countdown == 0:
                    break
                countdown -= 1

            # Join back successful threads otherwise threads will be killed
            # when the application ends.
            if self.thread.run_flag.is_set():
                self.thread.join()
                print("Successfully terminated additional"
                      f" {self.integration_name} integration thread.")

        if self.invocation.uses_operator():
            # If kamel is running the cluster then use kamel delete to
            # terminate the integration. First we terminate any services
            # associated with the integration.
            if self.service_name is not None:
                if not kubernetes.delete_service(mode, self.service_name):
                    raise RuntimeError(
                        f'Service with name {self.service_name} for'
                        '{self.integration_name} could not be'
                        'terminated')

            # Terminate the integration itself.
            if not kamel.delete(self.invocation, self.integration_name):
                raise RuntimeError(
                    f'Failed to terminate {self.integration_name}')
            return

        # If no kamel operator is used the only other alternative is that the
        # integration is running locally. In that case we only need to kill the
        # process that runs it.
        self.invocation.kill()

    # Get any type restrictions:
    def get_restricted_data_type(self):
        # If there are no restrictions return immediately:
        return self.input_restrictions['restricted_message_types']

    # Method that checks if, based on the configuration, the integration
    # requires something to be run or created before the integration is run.
    def prepare_environment(self, mode):
        # Create a multi-partition topic for a kafka source/sink.
        if (self.config['kind'] == 'kafka-source' or
            self.config['kind'] == 'kafka-sink') and \
           'partitions' in self.config and self.config['partitions'] > 1:
            topic = self.config['topic']
            partitions = self.config['partitions']
            kafka_brokers = self.config['brokers']

            # Create topic
            create_partitioned_topic(topic, partitions, kafka_brokers)

        # Create a multi-partition topic for the Kafka transport of a
        # source/sink.
        if mode.transport == "kafka" and \
           'kafka_transport_partitions' in self.config and \
           self.config['kafka_transport_partitions'] > 1:
            partitions = self.config['kafka_transport_partitions']

            # Create topic
            create_partitioned_topic(self.kafka_transport_topic, partitions,
                                     brokers())

    def route(self, default=None):
        if 'route' in self.config and self.config['route'] is not None:
            return self.config['route']
        if default is not None:
            return default
        return f'/{self.stream_name}'

    def use_backend(self):
        if 'use_backend' in self.config and self.config[
                'use_backend'] is not None:
            return self.config['use_backend']
        return False



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ray
from ray import serve
from rayvens.core.kamel_backend import KamelBackend
from rayvens.core import utils
from rayvens.core import common
from rayvens.core.catalog import construct_source, construct_sink
from rayvens.core.common import await_start
from rayvens.core.integration import Integration


def start(camel_mode, check_port, release):
    return Camel(common.get_run_mode(camel_mode, check_port, release))


class Camel:
    def __init__(self, mode):
        self.mode = mode
        self.mode.transport = 'ray-serve'

        # The Ray Serve backend used for Sinks. Sink are a special case and
        # can use one backend to support multiple sinks.
        self.kamel_backend = None

        # Start server is using a backend.
        serve.start()

    def add_source(self, stream, source, source_name):
        # Construct integration
        integration = Integration(stream.name, source_name, source)

        # Construct endpoint. First, get route:
        route = integration.route()

        # Determine the `to` endpoint value made up of a base address and
        # a custom route provided by the user. The computation depends on
        # the connector type used for the implementation.
        server_pod_name = ""
        if self.mode.is_cluster():
            server_pod_name = utils.get_server_pod_name()
        endpoint_base = self.mode._get_server_address(server_pod_name,
                                                      serve_source=True)

        # Construct integration source code. When the ray serve connector is
        # not enabled, use an HTTP inverted connection.
        integration_content = construct_source(source,
                                               f'{endpoint_base}{route}')

        # Set endpoint and integration names.
        endpoint_name = self._get_endpoint_name(integration.integration_name)

        # Create backend for this topic.
        source_backend = KamelBackend(self.mode, topic=stream.actor)

        # Create endpoint.
        source_backend.createProxyEndpoint(endpoint_name, route,
                                           integration.integration_name)

        # Start running the source integration.
        integration.invoke_run(self.mode, integration_content)

        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start source')
        return integration

    def add_sink(self, stream, sink, sink_name):
        # Construct integration
        integration = Integration(stream.name, sink_name, sink)

        # Extract integration properties:
        route = integration.route()
        use_backend = integration.use_backend()
        integration_name = integration.integration_name

        # Get integration source code.
        integration_content = construct_sink(sink, f'platform-http:{route}')

        # Create backend if one hasn't been created so far.
        if use_backend and self.kamel_backend is None:
            self.kamel_backend = KamelBackend(self.mode)

        # Start running the integration.
        integration.invoke_run(self.mode, integration_content)

        if use_backend:
            endpoint_name = self._get_endpoint_name(stream.name)
            self.kamel_backend.createProxyEndpoint(endpoint_name, route,
                                                   integration_name)

            helper = HelperWithBackend.remote(self.kamel_backend,
                                              serve.get_handle(endpoint_name),
                                              endpoint_name)
        else:
            helper = common.ProducerActor.remote(
                self.mode.server_address(integration_name) + route)
        stream.actor.send_to.remote(helper, sink_name)

        # Wait for integration to finish.
        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start sink')

        return integration

    def disconnect(self, integration):
        integration.disconnect(self.mode)

    def _get_endpoint_name(self, integration_name):
        return "_".join(["endpoint", integration_name])


@ray.remote(num_cpus=0)
class HelperWithBackend:
    def __init__(self, backend, endpoint_handle, endpoint_name):
        self.backend = backend
        self.endpoint_name = endpoint_name
        self.endpoint_handle = endpoint_handle

    def append(self, data):
        if data is not None:
            answer = self.backend.postToProxyEndpointHandle(
                self.endpoint_handle, self.endpoint_name, data)
            print(answer)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from rayvens.core.common import get_run_mode, send_to, await_start
from rayvens.core.common import ProducerActor
from rayvens.core.catalog import construct_source, construct_sink
from rayvens.core.integration import Integration


def start(camel_mode, check_port, release):
    return Camel(get_run_mode(camel_mode, check_port, release))


class Camel:
    def __init__(self, mode):
        self.mode = mode
        self.mode.transport = 'http'

    def add_source(self, stream, source, source_name):
        # Construct integration
        integration = Integration(stream.name, source_name, source)

        # Construct endpoint. First, get route:
        route = integration.route()

        # Prepare env:
        integration.prepare_environment(self.mode)

        # Determine the `to` endpoint value made up of a base address and
        # a custom route provided by the user. Use this to construct the
        # integration source code.
        integration_content = construct_source(source,
                                               f'platform-http:{route}',
                                               inverted=True)

        # Start running the source integration.
        integration.invoke_run(self.mode, integration_content)

        # Set up source for the HTTP connector case.
        integration.thread = send_to(stream.actor,
                                     self.mode.server_address(integration),
                                     route)

        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start source')
        return integration

    def add_sink(self, stream, sink, sink_name):
        # Construct integration
        integration = Integration(stream.name, sink_name, sink)

        # Extract integration properties:
        route = integration.route()

        # Prepare env:
        integration.prepare_environment(self.mode)

        # Get integration source code.
        integration_content = construct_sink(sink, f'platform-http:{route}')

        # Start running the integration.
        integration.invoke_run(self.mode, integration_content)

        helper = ProducerActor.remote(
            self.mode.server_address(integration) + route)
        stream.actor.send_to.remote(helper, sink_name)

        # Wait for integration to finish.
        if not await_start(self.mode, integration):
            raise RuntimeError('Could not start sink')

        return integration

    def disconnect(self, integration):
        integration.disconnect(self.mode)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from enum import Enum
from rayvens.core import invocation


class KubectlCommand(Enum):
    GET_PODS = 1
    GET_SERVICES = 2
    GET_DEPLOYMENTS = 3
    APPLY = 4
    DELETE_SERVICE = 5
    LOGS = 6


def kubectl_command_type(command):
    if command.startswith("get pods"):
        return KubectlCommand.GET_PODS
    if command.startswith("get services"):
        return KubectlCommand.GET_SERVICES
    if command.startswith("get deployments"):
        return KubectlCommand.GET_DEPLOYMENTS
    if command.startswith("apply"):
        return KubectlCommand.APPLY
    if command.startswith("delete service"):
        return KubectlCommand.DELETE_SERVICE
    if command.startswith("logs"):
        return KubectlCommand.LOGS
    raise RuntimeError('unsupported kubectl subcommand: %s' % command)


def kubectl_command_str(command_type):
    if command_type == KubectlCommand.GET_PODS:
        return "get pods"
    if command_type == KubectlCommand.GET_SERVICES:
        return "get services"
    if command_type == KubectlCommand.GET_DEPLOYMENTS:
        return "get deployments"
    if command_type == KubectlCommand.APPLY:
        return "apply"
    if command_type == KubectlCommand.DELETE_SERVICE:
        return "delete service"
    if command_type == KubectlCommand.LOGS:
        return "logs"
    raise RuntimeError('unsupported kubectl subcommand')


def kamel_command_end_condition(subcommand_type, service_name):
    if subcommand_type == KubectlCommand.GET_PODS:
        return ""
    if subcommand_type == KubectlCommand.GET_SERVICES:
        return ""
    if subcommand_type == KubectlCommand.GET_DEPLOYMENTS:
        return ""
    if subcommand_type == KubectlCommand.APPLY:
        if service_name is None or service_name == "":
            raise RuntimeError("Invalid kubernetes service name.")
        return "%s created" % service_name
    if subcommand_type == KubectlCommand.DELETE_SERVICE:
        if service_name is None or service_name == "":
            raise RuntimeError("Invalid kubernetes service name.")
        return "service \"%s\" deleted" % service_name
    if subcommand_type == KubectlCommand.LOGS:
        return "Installed features:"
    raise RuntimeError('unsupported kubectl subcommand: %s' %
                       kubectl_command_str(subcommand_type))


def extract_pod_name(line, pod_base_name):
    # Check if line contins pod name.
    if pod_base_name not in line:
        return ""

    # Split line into tokens.
    words = line.split()
    # If line is too short to be a valid pod line then exit.
    if len(words) < 1:
        return ""

    return words[0]


def service_name_matches(line, service_name):
    # Check if line contins pod name.
    if service_name not in line:
        return False

    # Split line into tokens.
    words = line.split()
    # If line is too short to be a valid pod line then exit.
    if len(words) < 1:
        return False

    return words[0] == service_name


def pod_state(line, pod_name, states):
    if pod_name == "" or line == "":
        return False

    # Split line into tokens.
    words = line.split()
    # If line is too short to be a valid pod line then exit.
    if len(words) < 2:
        return False

    if words[0] == pod_name:
        for state in states:
            if state == words[2]:
                return True
    return False


def is_pod_state_running(line, pod_name):
    return pod_state(line, pod_name, ["Running"])


def is_pod_state_error(line, pod_name):
    return pod_state(line, pod_name, ["Error", "CrashLoopBackOff"])


# Helper for checking pod status.
def pod_status(command, integration_name):
    # Invoke command using the Kubectl invocation class.
    kubectl_invocation = invocation.KubectlInvocation(command)

    # Wait for kubectl command to finish checking the integration.
    pod_running = kubectl_invocation.pod_is_running(integration_name)

    # Stop kubectl command.
    kubectl_invocation.kill()

    # Return pod status and full name.
    return pod_running, kubectl_invocation.pod_name


# Helper for running a kubectl command.
def invoke_kubectl_command(command,
                           message=None,
                           service_name=None,
                           with_output=False,
                           ongoing=False):
    # Invoke command using the Kubectl invocation class.
    kubectl_invocation = invocation.KubectlInvocation(command, service_name)

    # Wait for kubectl command to return.
    outcome = kubectl_invocation.invoke(message,
                                        service_name,
                                        with_output=with_output)

    # Stop kubectl ongoing command.
    if ongoing:
        kubectl_invocation.kill()

    # Return outcome.
    return outcome



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import random
import threading
from queue import Queue, Empty

# Port externalized by the cluster.
externalized_cluster_port = "31095"

# Check if the executable exists in PATH. This method should work
# in Windows, Linux and MacOS. Python >= 3.3 required.

rayvens_random = random.Random()
rayvens_random.seed()


def random_port(check_port=True, start=49152, end=65535):
    port = rayvens_random.randint(start, end)
    if not check_port:
        return port

    while not _port_is_free(port):
        print(f"Port {port} busy, trying new port.")
        port = rayvens_random.randint(start, end)

    return port


def _port_is_free(port):
    import socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Attempt to bind to the port if it fails, port must be busy.
    port_is_free = True
    try:
        sock.bind(('127.0.0.1', port))
    except socket.error as message:
        print(f'Port {port} is in use (Error: {str(message)}).')
        port_is_free = False

    sock.close()
    return port_is_free


def executable_is_available(executable):
    # If this is a path to the executable file return true.
    if os.path.isfile(executable):
        return True

    # Look up executable in path.
    from shutil import which
    return which(executable) is not None


def subprocess_tag(subprocess_name):
    return "[%s subprocess]" % subprocess_name


def print_log_from_subprocess(subprocess_name, stdout, with_output):
    output = stdout.readline().decode("utf-8")
    output = output.strip()

    if output != "" and with_output:
        print(subprocess_tag(subprocess_name), output)

    return output


class LogThread(threading.Thread):
    def __init__(self, stdout):
        threading.Thread.__init__(self)
        self.stop_flag = threading.Event()
        self.read_flag = threading.Event()
        self.stdout = stdout
        self.queue = Queue()

    def run(self):
        while not self.stop_flag.is_set():
            if self.read_flag.is_set():
                line = self.stdout.readline().decode("utf-8").strip()
                if line != "":
                    self.queue.put(line)
                    self.read_flag.clear()

        print("[Logging thread] Kamel command logging terminated.")


def print_log_from_queue(subprocess_name, queue, with_output):
    try:
        line = queue.get_nowait()
    except Empty:
        return None
    else:
        if with_output:
            print(subprocess_tag(subprocess_name), line)
        return line


def print_log(subprocess_name, message):
    print(subprocess_tag(subprocess_name), message)


def get_server_pod_name():
    with open('/etc/podinfo/labels', 'r') as f:
        for line in f:
            k, v = line.partition('=')[::2]
            if k == 'component':
                return f'{v[1:-2]}'

    raise RuntimeError("Cannot find server pod name")


def create_partitioned_topic(topic, partitions, brokers):
    # Create new topic
    from confluent_kafka.admin import AdminClient, NewTopic
    admin_client = AdminClient({"bootstrap.servers": brokers})

    # TODO: Smart choice for replication factor, for now use a
    # replication factor of 1.
    topics = [NewTopic(topic, num_partitions=partitions, replication_factor=1)]
    admin_client.create_topics(topics)

    # Wait for topic to be ready:
    topic_ready = False
    while not topic_ready:
        for enabled_topic in admin_client.list_topics().topics:
            if topic == enabled_topic:
                topic_ready = True
                break
    print(f"Topic {topic} is ready.")



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import subprocess
import rayvens.cli.utils as utils
from rayvens.cli.utils import PRINT

kamel_tag = "kamel"


def kamel_local_build_base_image(args):
    command = ["kamel"]

    # Local command:
    command.append("local")

    # Build command:
    command.append("build")

    # Base image options:
    command.append("--base-image")

    # Registry:
    registry = utils.get_registry(args)
    command.append("--container-registry")
    command.append(registry)

    # Wait for docker command to finish before returning:
    outcome = subprocess.run(command)

    image_name = utils.get_base_image_name(args)

    if outcome.returncode == 0:
        PRINT(f"Base image {image_name} pushed successfully.", tag=kamel_tag)
    else:
        PRINT(f"Base image {image_name} push failed.", tag=kamel_tag)


def kamel_local_build_image(args, integration_file_path):
    command = ["kamel"]

    # Local command:
    command.append("local")

    # Build command:
    command.append("build")

    # Image:
    integration_image = utils.get_integration_image(args)
    command.append("--image")
    command.append(integration_image)

    # Add integration file:
    command.append(integration_file_path)

    # Wait for docker command to finish before returning:
    outcome = subprocess.run(command)

    if outcome.returncode == 0:
        PRINT(f"Base image {integration_image} pushed successfully.",
              tag=kamel_tag)
    else:
        PRINT(f"Base image {integration_image} push failed.", tag=kamel_tag)



#
# Copyright IBM Corporation 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import yaml
import rayvens.cli.file as file
import rayvens.cli.java as java
from rayvens.core.catalog_utils import get_all_properties
from rayvens.core.catalog_utils import integration_requirements
from rayvens.core.catalog_utils import get_modeline_properties
from rayvens.core.catalog_utils import fill_config
from rayvens.core import catalog_utils

base_image_name = "integration-base"
kube_proxy_image_name = "kube-proxy"
property_prefix = "property: "
envvar_prefix = "envvar: "
job_launcher_service_account = "job-launcher-service-account"
job_launcher_cluster_role_binding = "job-launcher-service-account"
job_manager_role = "job-manager-role"
verbose = False


def get_summary_file(args):
    integration_summary = file.SummaryFile()

    integration_summary.kind = args.kind
    integration_summary.launch_image = args.launch_image

    if args.properties is not None:
        config, missing_property_value = fill_config(args.kind,
                                                     args.properties,
                                                     show_missing=False)
        for key in config:
            if key not in missing_property_value:
                integration_summary.add_property(key, config[key])

    envvars = get_current_envvars(args)
    for property_key in envvars:
        integration_summary.add_envvar(property_key, envvars[property_key])

    return integration_summary


def _create_file(workspace_directory, file_name, file_contents):
    file_processor_file_path = workspace_directory.joinpath(file_name)
    with open(file_processor_file_path, 'w') as f:
        f.write(file_contents)
    return file_name


def get_additional_files(spec, inverted_transport, launch_image):
    files = []
    if catalog_utils.integration_requires_file_processor(spec):
        files.append(
            file.File("ProcessFile.java",
                      contents=java.get_process_file_contents()))

    if catalog_utils.integration_requires_path_processor(spec):
        files.append(
            file.File("ProcessPath.java",
                      contents=java.get_process_path_contents()))

    # Write the Java queue code to the file when using HTTP transport.
    if inverted_transport:
        if catalog_utils.integration_requires_file_queue(spec):
            files.append(
                file.File("FileQueue.java",
                          contents=java.get_java_file_queue_contents()))

        if catalog_utils.integration_requires_file_queue(spec):
            files.append(
                file.File("FileQueueJson.java",
                          contents=java.get_java_file_queue_json_contents()))

        if catalog_utils.integration_requires_file_watch_queue(spec):
            files.append(
                file.File("FileWatchQueue.java",
                          contents=java.get_java_file_watch_queue_contents()))

        if catalog_utils.integration_requires_meta_event_queue(spec):
            files.append(
                file.File("MetaEventQueue.java",
                          contents=java.get_java_meta_event_queue_contents()))

        if catalog_utils.integration_requires_queue(spec):
            files.append(
                file.File("Queue.java",
                          contents=java.get_java_queue_contents(launch_image)))
    return files


def get_registry(args):
    if args.dev:
        return "localhost:5000"
    return args.registry


def get_given_properties(args):
    given_properties = []
    if args.properties is not None and len(args.properties) > 0:
        for property_value in args.properties:
            components = property_value.split("=")
            given_properties.append(components[0])
    return given_properties


def get_given_property_envvars(args):
    given_envvars = []
    if args.envvars is not None and len(args.envvars) > 0:
        for property_value in args.envvars:
            components = property_value.split("=")
            given_envvars.append(components[0])
    return given_envvars


def get_given_envvars(args):
    given_envvars = []
    if args.envvars is not None and len(args.envvars) > 0:
        for property_value in args.envvars:
            components = property_value.split("=")
            given_envvars.append("=".join(components[1:]))
    return given_envvars


def check_properties(kind, properties):
    invalid_props = []
    valid_properties = get_all_properties(kind)
    for property_name in properties:
        if property_name not in valid_properties:
            invalid_props.append(property_name)
    return invalid_props


def summary_file_path(workspace_directory):
    return workspace_directory.joinpath("summary.txt")


def _get_field_from_summary(summary_file_path, field, prefix=None):
    result = None
    with open(summary_file_path, "r") as summary:
        for line in summary.readlines():
            if prefix is not None:
                if not line.startswith(prefix):
                    continue
                line = line[len(prefix):]

            components = line.split("=")
            if components[0] == field:
                result = "=".join(components[1:])
                break

    if result is None:
        return None

    return result.strip()


def summary_get_kind(workspace_directory):
    summary_path = summary_file_path(workspace_directory)
    return _get_field_from_summary(summary_path, "kind")


def summary_get_envvar_properties(kind, summary_file, given_envvars):
    envvars = []
    valid_properties = get_all_properties(kind)
    for property_name in valid_properties:
        if property_name not in given_envvars and \
           property_name in summary_file.envvars:
            property_value = summary_file.envvars[property_name]
            if property_value is not None:
                envvars.append(f"{property_name}={property_value}")
    return envvars


def summary_get_envvars(kind, workspace_directory):
    envvars = []
    summary_path = summary_file_path(workspace_directory)
    valid_properties = get_all_properties(kind)
    for property_name in valid_properties:
        property_value = _get_field_from_summary(summary_path,
                                                 property_name,
                                                 prefix=envvar_prefix)
        if property_value is not None:
            envvars.append(property_value)
    return envvars


def summary_get_properties(kind, summary_file, given_properties):
    properties = []
    valid_properties = get_all_properties(kind)
    for property_name in valid_properties:
        if property_name not in given_properties and \
           property_name in summary_file.properties:
            property_value = summary_file.properties[property_name]
            if property_value is not None:
                properties.append(f"{property_name}={property_value}")
    return properties


def get_current_envvars(args):
    envvars = {}
    if args.envvars is not None and len(args.envvars) > 0:
        for property_env_pair in args.envvars:
            components = property_env_pair.split("=")
            envvars[components[0]] = "=".join(components[1:])
    return envvars


def get_current_config(args):
    requirements = integration_requirements(args.kind)

    config = dict(kind=args.kind)

    # Fill in properties if any have been provided.
    missing_requirements = []
    if args.properties is not None and len(args.properties) > 0:
        config, _ = fill_config(args.kind, args.properties, show_missing=False)

    # Fill in environment-based properties if any have been provided.
    if args.envvars is not None and len(args.envvars) > 0:
        for property_env_pair in args.envvars:
            components = property_env_pair.split("=")
            config[components[0]] = "=".join(components[1:])

    if len(requirements['required']) > 0:
        for req_property in requirements['required']:
            if req_property not in config:
                missing_requirements.append(req_property)
                config[req_property] = "missing_property_value"

    return config, missing_requirements


def get_full_config(summary_file, args):
    # Get the kind of the integration:
    kind = summary_file.kind

    # Get properties given as args:
    given_properties = get_given_properties(args)

    # Validate user-given properties:
    invalid_props = check_properties(kind, given_properties)
    if len(invalid_props) > 0:
        invalid_props = " ".join(invalid_props)
        raise RuntimeError(f"Invalid properties provided: {invalid_props}")

    # Assemble list of all property-value pairs:
    properties = []
    if args.properties is not None:
        properties = args.properties
    properties.extend(
        summary_get_properties(kind, summary_file, given_properties))

    # Fill configuration with values:
    config, _ = fill_config(kind, properties, show_missing=False)
    return config


def get_modeline_config(args, summary_file=None):
    # Get the kind of the integration:
    if summary_file is not None:
        kind = summary_file.kind
    else:
        kind = args.kind

    # Get envvars given as args:
    given_envvars = get_given_property_envvars(args)

    # Validate user-given envvars:
    invalid_props = check_properties(kind, given_envvars)
    if len(invalid_props) > 0:
        invalid_props = " ".join(invalid_props)
        raise RuntimeError(f"Invalid properties provided: {invalid_props}")

    # Assemble list of all property-value envvar pairs:
    envvars = []
    if args.envvars is not None:
        envvars = args.envvars
    if summary_file is not None:
        envvars.extend(
            summary_get_envvar_properties(kind, summary_file, given_envvars))

    # Transform configuarion in list of modeline properties:
    modeline_properties = get_modeline_properties(kind, envvars)
    result = []
    for key in modeline_properties:
        result.append(modeline_properties[key])
    return "\n".join(result)


def get_modeline_envvars(summary_file, args):
    # Get envvars given as args:
    given_envvars = get_given_envvars(args)

    # Get envvars from summary file:
    given_envvars.extend(summary_file.get_envvars())
    return given_envvars


def get_kubernetes_deployment_file_name(name):
    return f"{name}-deployment.yaml"


def get_kubernetes_integration_name(name):
    return f"{name}-integration"


def get_kubernetes_integration_file_name(name):
    return f"{name}-integration.yaml"


def get_kubernetes_entrypoint_name(name):
    return f"{name}-entrypoint"


def get_kubernetes_label_name(name):
    return f"{name}-label"


def get_base_image_name(args):
    # Registry name:
    registry = get_registry(args)

    # Base image name:
    return registry + "/" + base_image_name


def get_integration_image(args):
    # Registry name:
    registry = get_registry(args)

    # Actual image name:
    image_name = None
    if hasattr(args, "kind"):
        image_name = args.kind
    if hasattr(args, "image") and args.image is not None:
        image_name = args.image

    # When registry is not provided the image name on the command line
    # should contain the full path to the image.
    if registry is None:
        return args.image

    # Integration image name:
    return registry + "/" + image_name


def extract_image_name(args):
    if args.image is None:
        raise RuntimeError("Missing image name")
    registry = get_registry(args)
    if registry is None:
        components = args.image.split("/")
        return components[-1]
    return args.image


def PRINT(output, tag=None):
    if verbose:
        tag_str = ""
        if tag is not None:
            if not isinstance(tag, str):
                raise RuntimeError("Invalid output tag.")
            tag_str = "[" + tag + "]"
        print(tag_str, output)


def VERBOSE_MODE():
    return verbose
